import numpy as np
from pyscipopt import Model
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.tree import DecisionTreeRegressor
from utils import read_csv_to_dict, train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of a city manager in Poland.

We have access to open source data
(thanks to: https://www.kaggle.com/datasets/krzysztofjamroz/apartment-prices-in-poland),
for which we have built a predictor. The predictor takes as input a variety of apartment features,
and outputs the price of the apartment, which we consider as a proxy for living quality.

The goal of this MIP is to maximise the living quality of the randomly sampled 50 residents by
building two new schools, clinics, post offices, kindergartens, restaurants,
and pharmacies, as well as a single college. For the sake simplicity we assume the
closest access for all these residents to all these facilities is over 10km away.
To further simplify the MIP we measure distance via the L1 norm at all times.

Let f to be the ML predictor. Let I be the index set of residents and their apartments.
Let J be the index of the building type and nJ be the amount of that building type that will be built.
Let k be the index set of the imaginary 2D grid that will be used for L1 distance.

Let a_loc[i][k] be the location of the apartment i in dimension k

Let build[J][nJ][k] be the variables representing the location of the buildings.
Let x[I][J] be the min distance between the buildings and apartments
Let y[i] be the price of the apartment

The MIP we model below in a city of size 5kmx5km is:

For all building types where two are built ensure a minimum distance between each:
    build[J][0][k] - build[J][1][k] == pos_slack[J][k] - neg_slack[J][k] for all J where nJ == 2 for all k
    pos_slack[J][k] <= 10 * bin_slack[J][k] for all J where nJ == 2 for all k
    neg_slack[J][k] <= 10 * (1 - bin_slack[J][k]) for all J where nJ == 2 for all k
    sum_k pos_slack[J][k] + neg_slack[J][k] >= 1 for all J where nJ == 2

For all apartments track the minimum distance between each building

build[j][nJ][k] - a_loc[i][k] == pos_slack[i][j][nJ][k] - neg_slack[i][j][nJ][k] for all I, J, nJ, k
pos_slack[i][j][nJ][k] <= 10 * bin_slack[i][j][nJ][k] for all I, J, nJ, k
neg_slack[i][j][nJ][k] <= 10 * (1 - bin_slack[i][j][nJ][k]) for all I, J, nJ, k
dist_a_b_k[i][j][nJ][k] = pos_slack[i][j][nJ][k] + neg_slack[i][j][nJ][k] for all I, J, nJ, k
dist_a_b[i][j][nJ] = sum_k dist_a_b_k[i][j][nJ][k] for all I, J, nJ

if nJ == 1
x[i][j] = dist_a_b[i][j][0]

if nJ == 2
dist_a_b[i][j][1] - dist_a_b[i][j][0] <= 10 * bin_slack[i][j] for all I, J
dist_a_b[i][j][0] - dist_a_b[i][j][1] <= 10 * (1 - bin_slack[i][j]) for all I, J
x[i][j] <= dist_a_b[i][j][0] for all I, J
x[i][j] <= dist_a_b[i][j][1] for all I, J
x[i][j] >= dist_a_b[i][j][0] - 10 * (1 - bin_slack[i][j]) for all I, J
x[i][j] >= dist_a_b[i][j][1] - 10 bin_slack[i][j] for all I, J

The embedded ML constraints also need to be added
y[i] = f(x[i][:]) for all I

max(sum_i y[i])
"""


def build_and_optimise_city_manager(
    data_seed=42,
    training_seed=42,
    dt_gbdt_or_mlp="dt",
    formulation="sos",
    framework="sklearn",
    max_depth_or_layer_size=6,
    n_estimators_layers=3,
    n_apartments=25,
    epsilon=0.001,
    build_only=False,
):
    # Path to apartment price data
    data_dict = read_csv_to_dict("./tests/data/apartments.csv")
    grid_length = 5

    # Set the random seed
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)

    # The features of our predictor. All distance based features are variables.
    features = [
        "squareMeters",
        "rooms",
        "floorCount",
        "centreDistance",
        "schoolDistance",
        "clinicDistance",
        "postOfficeDistance",
        "kindergartenDistance",
        "restaurantDistance",
        "collegeDistance",
        "pharmacyDistance",
        "hasParkingSpace",
        "hasBalcony",
        "hasSecurity",
    ]
    feature_to_idx = {feature: i for i, feature in enumerate(features)}
    n_features = len(features)

    # Generate the actual input data arrays for the ML predictors
    X = []
    y = np.array([float(x) / 10**5 for x in data_dict["price"]]).reshape(-1, 1)
    for feature in features:
        if "has" in feature:
            X.append(np.array([x == "yes" for x in data_dict[feature]]))
        else:
            X.append(np.array([float(x) for x in data_dict[feature]]))
    X = np.swapaxes(np.array(X), 0, 1)

    # Train the ML predictor
    if dt_gbdt_or_mlp == "dt":
        reg = DecisionTreeRegressor(
            random_state=training_random_state, max_depth=max_depth_or_layer_size
        ).fit(X, y)
    elif dt_gbdt_or_mlp == "gbdt":
        reg = GradientBoostingRegressor(
            random_state=training_random_state,
            max_depth=max_depth_or_layer_size,
            n_estimators=n_estimators_layers,
        ).fit(X, y.reshape(-1))
    elif dt_gbdt_or_mlp == "mlp":
        if framework == "sklearn":
            hidden_layers = tuple([max_depth_or_layer_size for i in range(n_estimators_layers)])
            reg = MLPRegressor(
                random_state=training_random_state, hidden_layer_sizes=hidden_layers
            ).fit(X, y.reshape(-1))
        else:
            reg = train_torch_neural_network(
                X, y, n_estimators_layers, max_depth_or_layer_size, training_seed, reshape=False
            )
    else:
        raise ValueError(f"Unknown value: {dt_gbdt_or_mlp}")

    # Create the SCIP Model
    scip = Model()

    # Create variables for the n_apartments many imaginary apartments.The apartments lie on an imaginary square that is
    # defined by the four corners [(0, 0), (0, grid_length), (grid_length, grid_length), (grid_length, 0)]
    feature_vars = np.zeros((n_apartments, n_features), dtype=object)
    price_vars = np.zeros((n_apartments, 1), dtype=object)
    apartment_locations = data_random_state.uniform(0, grid_length, size=(n_apartments, 2))
    rooms = np.random.randint(
        np.min(X[feature_to_idx["rooms"]]),
        np.max(X[feature_to_idx["rooms"]]) + 1,
        size=n_apartments,
    )
    floors = np.random.randint(
        np.min(X[feature_to_idx["floorCount"]]),
        np.max(X[feature_to_idx["floorCount"]]) + 1,
        size=n_apartments,
    )
    square_meters = np.random.uniform(
        np.min(X[feature_to_idx["squareMeters"]]),
        np.max(X[feature_to_idx["squareMeters"]]),
        size=n_apartments,
    )

    # Create variables for where the various buildings will be placed
    school_vars = np.zeros((2, 2), dtype=object)
    clinic_vars = np.zeros((2, 2), dtype=object)
    post_office_vars = np.zeros((2, 2), dtype=object)
    kindergarten_vars = np.zeros((2, 2), dtype=object)
    restaurant_vars = np.zeros((2, 2), dtype=object)
    college_vars = np.zeros((1, 2), dtype=object)
    pharmacy_vars = np.zeros((2, 2), dtype=object)
    building_vars = [
        ("school", "schoolDistance", school_vars),
        ("clinic", "clinicDistance", clinic_vars),
        ("post_office", "postOfficeDistance", post_office_vars),
        ("kindergarten", "kindergartenDistance", kindergarten_vars),
        ("restaurant", "restaurantDistance", restaurant_vars),
        ("college", "collegeDistance", college_vars),
        ("pharmacy", "pharmacyDistance", pharmacy_vars),
    ]

    # Now fill in the actual variables
    for i in range(n_apartments):
        price_vars[i][0] = scip.addVar(vtype="C", lb=-25, ub=100, name=f"price_{i}")
        for j, feature in enumerate(features):
            feature_vars[i][j] = scip.addVar(vtype="C", lb=0, name=f"feature_{i}_{j}")
            # Randomly generate characteristics
            if "has" in feature:
                scip.fixVar(feature_vars[i][j], int(np.random.randint(0, 2)))
            elif feature == "rooms":
                scip.fixVar(feature_vars[i][j], int(rooms[i]))
            elif feature == "floorCount":
                scip.fixVar(feature_vars[i][j], int(floors[i]))
            elif feature == "squareMeters":
                scip.fixVar(feature_vars[i][j], int(square_meters[i]))
            elif feature == "centreDistance":
                scip.fixVar(
                    feature_vars[i][j],
                    np.sum(
                        np.abs(
                            apartment_locations[i] - np.array([grid_length / 2, grid_length / 2])
                        )
                    ),
                )
            else:
                scip.chgVarUb(feature_vars[i][j], 2 * grid_length)

    for building_type, _, building_var in building_vars:
        for i in range(building_var.shape[0]):
            building_var[i][0] = scip.addVar(
                vtype="C", lb=0, ub=grid_length, name=f"{building_type}_{i}_0"
            )
            building_var[i][1] = scip.addVar(
                vtype="C", lb=0, ub=grid_length, name=f"{building_type}_{i}_1"
            )

    # Add constraints that link the distance of the built structures to each apartment
    for building_type, feature, building_var in building_vars:
        feature_idx = feature_to_idx[feature]
        n_buildings = building_var.shape[0]
        big_m = grid_length * 2

        # If there are two buildings ensure that there is a minimum distance between them
        if n_buildings == 2:
            pos_dists = [
                scip.addVar(vtype="C", lb=0, ub=big_m, name=f"pos__{building_type}_{i}_{k}")
                for k in range(2)
            ]
            neg_dists = [
                scip.addVar(vtype="C", lb=0, ub=big_m, name=f"neg_{building_type}_{i}_{k}")
                for k in range(2)
            ]
            bin_dists = [
                scip.addVar(vtype="B", name=f"bin_{building_type}_{i}_{k}") for k in range(2)
            ]
            for k in range(2):
                scip.addCons(
                    building_var[0][k] - building_var[1][k] == pos_dists[k] - neg_dists[k],
                    name=f"dist_build_{building_type}_{i}_{k}",
                )
                scip.addCons(
                    pos_dists[k] <= big_m * bin_dists[k],
                    name=f"big_m_pos_{building_type}_{i}_{k}",
                )
                scip.addCons(
                    neg_dists[k] <= big_m * (1 - bin_dists[k]),
                    name=f"big_m_neg_{building_type}_{i}_{k}",
                )
            scip.addCons(
                sum(pos_dists[k] + neg_dists[k] for k in range(2)) >= 0.8,
                name=f"min_dist_build_{building_type}",
            )

        # Add constraints that measure the distance between each apartment and the placed building
        for i in range(n_apartments):
            if building_var.shape[0] > 2:
                raise AssertionError(
                    "Didn't program formulations for more than 2 buildings of the same type"
                )

            dist_builds = [
                scip.addVar(
                    vtype="C", lb=0, ub=2 * grid_length, name=f"sum_dist_{building_type}_{i}_{j}"
                )
                for j in range(n_buildings)
            ]
            dist_dims = [
                [
                    scip.addVar(
                        vtype="C", lb=0, ub=big_m, name=f"dist_{building_type}_{i}_{j}_{k}"
                    )
                    for k in range(2)
                ]
                for j in range(n_buildings)
            ]
            pos_dists = [
                [
                    scip.addVar(
                        vtype="C", lb=0, ub=big_m, name=f"pos__{building_type}_{i}_{j}_{k}"
                    )
                    for k in range(2)
                ]
                for j in range(n_buildings)
            ]
            neg_dists = [
                [
                    scip.addVar(vtype="C", lb=0, ub=big_m, name=f"neg_{building_type}_{i}_{j}_{k}")
                    for k in range(2)
                ]
                for j in range(n_buildings)
            ]
            bin_dists = [
                [scip.addVar(vtype="B", name=f"bin_{building_type}_{i}_{j}_{k}") for k in range(2)]
                for j in range(n_buildings)
            ]
            for j in range(n_buildings):
                for k in range(2):
                    scip.addCons(
                        building_var[j][k] - apartment_locations[i][k]
                        == pos_dists[j][k] - neg_dists[j][k],
                        name=f"dist_apartment_{building_type}_{i}_{j}_{k}",
                    )
                    scip.addCons(
                        pos_dists[j][k] <= big_m * bin_dists[j][k],
                        name=f"big_m_pos_{building_type}_{i}_{j}_{k}",
                    )
                    scip.addCons(
                        neg_dists[j][k] <= big_m * (1 - bin_dists[j][k]),
                        name=f"big_m_neg_{building_type}_{i}_{j}_{k}",
                    )
                    scip.addCons(
                        dist_dims[j][k] == pos_dists[j][k] + neg_dists[j][k],
                        name=f"sum_dist_{building_type}_{i}_{j}_{k}",
                    )
                scip.addCons(
                    dist_builds[j] == dist_dims[j][0] + dist_dims[j][1],
                    name=f"sum_dist_dims_{i}_{building_type}",
                )
            if n_buildings == 1:
                scip.addCons(
                    feature_vars[i][feature_idx] == dist_builds[0],
                    name=f"feature_dist_{i}_{feature_idx}",
                )
            else:
                aux_max_var_bin = scip.addVar(vtype="B", name=f"aux_max_{i}_{building_type}")
                scip.addCons(
                    dist_builds[1] - dist_builds[0] <= big_m * aux_max_var_bin,
                    name=f"big_m_{i}_{building_type}_0",
                )
                scip.addCons(
                    dist_builds[0] - dist_builds[1] <= big_m * (1 - aux_max_var_bin),
                    name=f"big_m_{i}_{building_type}_1",
                )
                scip.addCons(
                    feature_vars[i][feature_idx] <= dist_builds[0],
                    name=f"dist_{i}_{building_type}_0",
                )
                scip.addCons(
                    feature_vars[i][feature_idx] <= dist_builds[1],
                    name=f"dist_{i}_{building_type}_1",
                )
                scip.addCons(
                    feature_vars[i][feature_idx] >= dist_builds[0] - big_m * (1 - aux_max_var_bin),
                    name=f"dist_{i}_{building_type}_2",
                )
                scip.addCons(
                    feature_vars[i][feature_idx] >= dist_builds[1] - big_m * aux_max_var_bin,
                    name=f"dist_{i}_{building_type}_3",
                )

    # Embed the ML predictors for each apartment. A small epsilon ensures we match sklearn
    pred_cons = add_predictor_constr(
        scip,
        reg,
        feature_vars,
        price_vars,
        epsilon=epsilon,
        formulation=formulation,
        unique_naming_prefix=f"predictor_",
    )

    # Add the objective to the MIP
    scip.setObjective(-np.sum(price_vars) + (20 * n_apartments))

    if not build_only:
        # Optimise the SCIP model
        scip.optimize()
        # We can check the "error" of the MIP embedding by determining the difference SKLearn and SCIP output
        if np.max(pred_cons.get_error()) > 10**-4:
            error = np.max(pred_cons.get_error())
            raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -4}")

    return scip


def test_dt_city_manager():
    build_and_optimise_city_manager(
        data_seed=42,
        training_seed=42,
        dt_gbdt_or_mlp="dt",
        max_depth_or_layer_size=4,
        n_estimators_layers=3,
        n_apartments=15,
        epsilon=0.001,
        build_only=False,
    )


def test_gbdt_city_manager():
    build_and_optimise_city_manager(
        data_seed=50,
        training_seed=80,
        dt_gbdt_or_mlp="gbdt",
        max_depth_or_layer_size=3,
        n_estimators_layers=3,
        n_apartments=15,
        epsilon=0.001,
        build_only=False,
    )


def test_mlp_city_manager():
    build_and_optimise_city_manager(
        data_seed=50,
        training_seed=80,
        dt_gbdt_or_mlp="mlp",
        formulation="sos",
        max_depth_or_layer_size=3,
        n_estimators_layers=2,
        n_apartments=2,
        epsilon=0.001,
        build_only=False,
    )


def test_mlp_bigm_city_manager():
    build_and_optimise_city_manager(
        data_seed=50,
        training_seed=80,
        dt_gbdt_or_mlp="mlp",
        formulation="bigm",
        max_depth_or_layer_size=3,
        n_estimators_layers=2,
        n_apartments=2,
        epsilon=0.001,
        build_only=False,
    )



""" Utilities for modelling gradient boosting decision trees and random forest constraints """

import numpy as np

from ...sklearn.decision_tree import (
    add_decision_tree_classifier_constr,
    add_decision_tree_regressor_constr,
)
from ..base_predictor_constraint import AbstractPredictorConstr
from ..classification.argmax_model import argmax_bound_formulation
from ..decision_tree import leaf_formulation


def aggregated_estimator_formulation(
    scip_model,
    _input,
    output,
    tree_vars,
    trees,
    constant,
    lr,
    n_estimators,
    unique_naming_prefix,
    epsilon,
    aggr,
    classification,
    **kwargs,
):
    """
    Creates the model that represents the aggregation of estimators into a single output.
    This function is used exclusively for the case where the estimators are decision trees, and the larger
    predictor is either a gradient boosting decision tree or random forest.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    _input : np.ndarray
        The input variables that are passed to each decision tree
    output : np.ndarray
        The output variables of the predictor
    tree_vars : np.ndarray
        The PySCIPOpt variables that have been created to represent the output of each decision tree (i.e. estimator)
    trees : list
        A list of lists containing dictionary information that completely describe each decision tree (i.e. estimator)
    constant : np.ndarray
        An array of constant shift values that are added to the output values of each decision tree (i.e. estimator)
    lr : float or int
        The learning rate used while training. For GBDT / RF this scales the output of each tree
    n_estimators : int
        The number of decision trees (i.e. estimators)
    unique_naming_prefix : str
        The unique naming prefix string that goes before all variables and constraints that are constructed by SCIP
    epsilon : float
        The epsilon that is used for each decision tree model. See
        :py:func:`pyscipopt_ml.modelling.decision_tree.leaf_formulation`.
    aggr : str, "sum" or "avg"
        The aggregation method used in the formulation. Either the estimators are averages or summed.
    classification : bool
        Whether the aggregated output of each decision tree (i.e. estimator) should be used for classification.

    Returns
    -------

    estimators : list
        A list of :py:class`pyscipopt_ml.modelling.aggregate_tree_model.TreeEstimator`
    created_vars : list
        A list containing all created PySCIPOpt vars
    created_cons : list
        A list containing all created PySCIPOpt cons
    """

    # Get the number of samples and output dimension
    n_samples = _input.shape[0]
    outdim = output.shape[-1]

    # Create the individual tree estimators
    estimators = create_tree_estimators(
        scip_model,
        _input,
        tree_vars,
        trees,
        n_estimators,
        outdim,
        unique_naming_prefix,
        epsilon,
        False,
        **kwargs,
    )

    # Aggregate the trees over the output dimension
    aggregate_tree_output = aggregate_estimator_outputs(tree_vars, lr, constant, aggr=aggr)

    # Formulate the appropriate constraints
    created_vars, created_cons = create_aggregation_constraints(
        scip_model,
        aggregate_tree_output,
        output,
        n_samples,
        outdim,
        unique_naming_prefix,
        classification,
    )

    return estimators, created_vars, created_cons


def create_aggregation_constraints(
    scip_model,
    aggregate_tree_output,
    output,
    n_samples,
    outdim,
    unique_naming_prefix,
    classification,
):
    """
    Creates the variables and constraints that link the output of the predictor itself and the aggregation of each
    estimator.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    aggregate_tree_output : np.ndarray
        The aggregated output variables of each decision tree
    output : np.ndarray
        The output variables of the predictor
    n_samples : int
        The number of samples
    outdim : int
        The number of outputs of each decision tree (i.e. estimator)
    unique_naming_prefix : str
        The unique naming prefix string that goes before all variables and constraints that are constructed by SCIP
    classification : bool
        Whether the aggregated output of each decision tree (i.e. estimator) should be used for classification.

    Returns
    -------
    created_vars : list
        A list containing all created PySCIPOpt vars
    created_cons : list
        A list containing all created PySCIPOpt cons
    """

    # Formulate the appropriate constraints
    created_cons = []
    created_vars = []
    if not classification:
        sum_tree_cons = np.zeros((n_samples, outdim), dtype=object)
        for i in range(n_samples):
            for j in range(outdim):
                name = unique_naming_prefix + f"tree_sum_{i}_{j}"
                sum_tree_cons[i][j] = scip_model.addCons(
                    output[i][j] == aggregate_tree_output[i][j], name=name
                )
        created_cons.append(sum_tree_cons)
    else:
        new_vars, new_cons = argmax_bound_formulation(
            scip_model, aggregate_tree_output, output, unique_naming_prefix, one_dim_center=0
        )  # TODO: Determine for which models is one_dim_center=0.5
        for added_var in new_vars:
            created_vars.append(added_var)
        for added_cons in new_cons:
            created_cons.append(added_cons)

    return created_vars, created_cons


def aggregate_estimator_outputs(_output, lr, constant, aggr="sum"):
    """
    Aggregate the output of individual estimators into a single expression for each output dimension.
    This function is needed for models with multiple estimators, e.g. gradient boosting decision trees and
    random forests.
    The output after aggregation can then be used as input for argmax classification.

    Parameters
    ----------
    _output : np.ndarray
        The output variables from each individual estimator (e.g. decision tree)
    lr : float
        The learning rate used for training and which is used to scale the output
    constant : np.ndarray
        The constant term that is added to the aggregation
    aggr : "sum" or "avg"
        Aggregation type ("sum" or "avg"). "Sum" for gradient boosting decision trees. "avg" for random forests.
    Returns
    -------

    aggregated_output : np.ndarray
        The new aggregated output per dimension over all estimators. Traditionally a sum over one dimension.

    """
    assert aggr in [
        "sum",
        "avg",
    ], f"Aggregation type {aggr} is neither sum or avg. No model exists."
    assert (
        _output.ndim == 3
    ), f"Aggregating estimator outputs of invalid dimension. {_output.ndim} != 3"

    n_samples = _output.shape[0]
    outdim = _output.shape[-1]
    n_estimators = _output.shape[1]

    aggregated_output = np.zeros((n_samples, outdim), dtype=object)
    for i in range(n_samples):
        for j in range(outdim):
            sum_expr = constant[j]
            for k in range(n_estimators):
                scale = 1 if aggr == "sum" else n_estimators
                sum_expr += lr * _output[i][k][j] / scale
            aggregated_output[i][j] = sum_expr

    return aggregated_output


def create_tree_estimators(
    scip_model,
    _input,
    tree_vars,
    trees,
    n_estimators,
    outdim,
    unique_naming_prefix,
    epsilon,
    classification,
    **kwargs,
):
    """
    Creates individual tree estimator models for each decision tree.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    _input : np.ndarray
        The input variables that are passed to each decision tree
    tree_vars : np.ndarray
        The PySCIPOpt variables that have been created to represent the output of each decision tree (i.e. estimator)
    trees : list
        A list of lists containing dictionary information that completely describe each decision tree (i.e. estimator)
    n_estimators : int
        The number of decision trees (i.e. estimators)
    outdim : int
        The output dimension of each decision tree
    unique_naming_prefix : str
        The unique naming prefix string that goes before all variables and constraints that are constructed by SCIP
    epsilon : float
        The epsilon that is used for each decision tree model. See #TODO: Decision tree modelling path
    classification : bool
        Whether the individual decision trees (i.e. estimators) are classification trees

    Returns
    -------

    estimators : list
        A list of :py:class`pyscipopt_ml.modelling.aggregate_tree_model.TreeEstimator`
    """

    estimators = []
    for i in range(n_estimators):
        for j in range(outdim):
            unique_prefix = unique_naming_prefix + f"{i}_{j}"
            estimators.append(
                TreeEstimator(
                    scip_model,
                    trees[i][j],
                    _input,
                    tree_vars[:, i, j].reshape((-1, 1)),
                    unique_prefix,
                    epsilon,
                    classification,
                    **kwargs,
                )
            )

    return estimators


def create_sklearn_tree_estimators(
    scip_model,
    predictor,
    _input,
    n_samples,
    outdim,
    unique_naming_prefix,
    classification,
    gbdt_or_rf="gbdt",
    **kwargs,
):
    """
    Create individual estimators for each decision tree for decision tree based ensemble predictors from SKLearn.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    predictor : GradientBoostingClassifier | GradientBoostingRegressor | RandomForestClassifier | RandomForestRegressor
        The Sklearn predictor that we are modelling
    _input : np.ndarray
        The input variables into each decision tree (i.e. estimator)
    n_samples : int
        The number of samples as input
    outdim : int
        The number of outputs of each estimator
    unique_naming_prefix : str
        The unique naming prefix string that goes before all variables and constraints that are constructed by SCIP
    classification : bool
        Whether the individual decision trees (i.e. estimators) are classification trees
    gbdt_or_rf : "gbdt" | "rf"
        Whether the predictor is for gradient boosting decision trees or random forests.

    Returns
    -------
    estimators : list
        A list of :py:class`pyscipopt_ml.modelling.aggregate_tree_model.TreeEstimator`
    tree_vars : np.ndarray
        A np.ndarray of created PySCIPopt vars
    """

    # Create variables to represent the output of each decision tree (i.e. estimator)
    tree_vars = scip_model.addMatrixVar(
        (n_samples, predictor.n_estimators, outdim),
        vtype="C",
        lb=None,
        name=unique_naming_prefix + "_tree_var",
    )

    # Create each estimator. In the case of GBDT, there are (n_estimators, outdim) many estimators, while for RF
    # there are (outdim,) many estimators. In the case of GBDT for classification each individual DT is regression.
    estimators = []
    if gbdt_or_rf == "gbdt":
        for i in range(predictor.n_estimators_):
            for j in range(outdim):
                unique_prefix = unique_naming_prefix + f"{i}_{j}"
                tree = predictor.estimators_[i][j]
                estimators.append(
                    add_decision_tree_regressor_constr(
                        scip_model,
                        tree,
                        _input,
                        tree_vars[:, i, j].reshape((-1, 1)),
                        unique_prefix,
                        **kwargs,
                    )
                )
    elif gbdt_or_rf == "rf":
        for i in range(predictor.n_estimators):
            tree = predictor.estimators_[i]
            unique_prefix = unique_naming_prefix + f"{i}"
            if classification:
                estimators.append(
                    add_decision_tree_classifier_constr(
                        scip_model, tree, _input, tree_vars[:, i, :], unique_prefix, **kwargs
                    )
                )
            else:
                estimators.append(
                    add_decision_tree_regressor_constr(
                        scip_model, tree, _input, tree_vars[:, i, :], unique_prefix, **kwargs
                    )
                )

    return estimators, tree_vars


class TreeEstimator(AbstractPredictorConstr):
    def __init__(
        self,
        scip_model,
        tree,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon,
        classification,
        **kwargs,
    ):
        self._tree = tree
        self._epsilon = epsilon
        self._classification = classification
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        new_vars, new_cons = leaf_formulation(
            self.scip_model,
            self.input,
            self.output,
            self._tree,
            self.unique_naming_prefix,
            self._epsilon,
            classification=self._classification,
        )

        for new_var in new_vars:
            self._created_vars.append(new_var)
        for new_con in new_cons:
            self._created_cons.append(new_con)

    def get_error(self, eps):
        raise AssertionError("Tree estimator has no get_error functionality")



"""Module for formulating a :external+sklearn:py:class:`sklearn.multioutput.MultiOutputRegressor`
or :external+sklearn:py:class:`sklearn.multioutput.MultiOutputClassifier`
into a PySCIPOpt Model.

"""

from ..exceptions import NoModel
from ..modelling.base_predictor_constraint import AbstractPredictorConstr
from ..modelling.get_convertor import get_convertor
from ..registered_predictors import sklearn_convertors
from .skgetter import SKgetter


def add_multi_output_regressor_constr(
    scip_model,
    multi_output_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate a multi_output_regressor into scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    multi_output_regressor.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the predictor should be inserted.
    multi_output_regressor : :external+sklearn:py:class:`sklearn.multioutput.MultiOutputRegressor`
        The multi_output_regressor to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for multi_output_regressor in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for multi_output_regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    MultiOutputConstr
        Object containing information about what was added to scip_model to embed the
        predictor into it

    Raises
    ------
    NoModel
        If the translation to SCIP of one of the elements in the multi_output_regressor
        is not implemented or recognized.

    Notes
    -----
    |VariablesDimensionsWarn|
    """
    return MultiOutputConstr(
        scip_model,
        multi_output_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification=False,
        **kwargs,
    )


def add_multi_output_classifier_constr(
    scip_model,
    multi_output_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate a multi_output_classifier into scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    multi_output_classifier.

    Parameters
    ----------
    scip_model : SCIP Model
        The PySCIPOpt Model where the predictor should be inserted.
    multi_output_classifier : :external+sklearn:py:class:`sklearn.multioutput.MultiOutputClassifier`
        The multi_output_classifier to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for multi_output_classifier in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for multi_output_classifier in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    MultiOutputConstr
        Object containing information about what was added to scip_model to embed the
        predictor into it

    Raises
    ------
    NoModel
        If the translation to SCIP of one of the elements in the multi_output_classifier
        is not implemented or recognized.

    Notes
    -----
    |VariablesDimensionsWarn|
    """
    return MultiOutputConstr(
        scip_model,
        multi_output_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification=True,
        **kwargs,
    )


class MultiOutputConstr(SKgetter, AbstractPredictorConstr):
    """Class to formulate a trained :external+sklearn:py:class:`sklearn.multioutput.MultiOutputRegressor`
    or :external+sklearn:py:class:`sklearn.multioutput.MultiOutputClassifier`
    into a PySCIPOpt model.

    |ClassShort|
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification,
        **kwargs,
    ):
        self._estimators = []
        self.classification = classification
        self.output_size = len(predictor.estimators_)
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            scip_model,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        for i, estimator in enumerate(self.predictor.estimators_):
            convertor = get_convertor(estimator, sklearn_convertors())
            if convertor is None:
                raise NoModel(
                    self.predictor,
                    f"The sklearn multi-output contains estimator {estimator}, which is not supported",
                )
            self._estimators.append(
                convertor(
                    self.scip_model,
                    estimator,
                    self.input,
                    self.output[:, i].reshape(-1, 1),
                    unique_naming_prefix=self.unique_naming_prefix + f"_step_{i}",
                    **kwargs,
                )
            )



""" Utilities for modelling argmax constraints """

import numpy as np
from pyscipopt import quicksum


def argmax_bound_formulation(scip_model, _input, output, unique_naming_prefix, one_dim_center=0.5):
    """
    Create constraints that represent the output of an argmax function applied to _input.
    The constraints ensure binary output of a single class.

    The formulation is different depending on the number of classes. In the case of there being two samples:

    Let x be the regression input \reals^{2}, and z the binary output {0, 1}^{2}
    .. math::

        \begin{align*}
        z_{1} : x_{1} >= x_{2}
        z_{2} : x_{2} >= x_{1}
        \sum z_{i} == 1
        \end{align*}

    for the case of arbitrary classes the formulation below is used:

    Let x be the regression input \reals^{n}, z the binary output
    {0, 1}^{n}, s the slack variables [0, inf]^{n}, and y the maximum over the input \reals:

    .. math::

        \begin{align*}
        x_{i} + s_{i} - y == 0 \forall i \in N
        SOS1(z_{i}, s_{i}) \forall i \in N
        \sum z_{i} == 1
        \end{align*}


    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    _input : np.ndarray
        The (potentially aggregated) output variables from the regression variant of a predictor, which are now
        input to the argmax formulation.
    output : np.ndarray
        The output variables of the (classification) predictor
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    one_dim_center : float, optional
        The value for which the 1-D argmax is centred around. Normally this is 0.5 for a single binary.
    Returns
    -------
    created_vars : list
        A list containing np.ndarray of PySCIPOpt variables that were created for the argmax formulation
    created_cons : list
        A list containing np.ndarray of PySCIPOpt constraints that we created for the argmax formulation
    """

    assert (
        _input.shape == output.shape
    ), f"Input and output dimensions do not match. {_input.shape} != {output.shape}"

    # get the in and out dimensions
    n_samples = _input.shape[0]
    outdim = output.shape[-1]

    # Separate the formulation into cases
    if outdim == 1:
        bin_vars = scip_model.addMatrixVar(
            (n_samples,), vtype="B", name=unique_naming_prefix + "_argmax"
        )

        # Create additional constraints
        output_equal_cons = np.zeros((n_samples,), dtype=object)
        output_under_half = np.zeros((n_samples,), dtype=object)
        output_over_half = np.zeros((n_samples,), dtype=object)

        # Now populate the constraints
        for i in range(n_samples):
            name = unique_naming_prefix + f"out_eq_{i}"
            output_equal_cons[i] = scip_model.addCons(output[i][0] == bin_vars[i], name=name)
            name = unique_naming_prefix + f"out_ub_{i}"
            output_under_half[i] = scip_model.addConsIndicator(
                _input[i][0] <= one_dim_center, bin_vars[i], activeone=False, name=name
            )
            name = unique_naming_prefix + f"out_lb_{i}"
            output_under_half[i] = scip_model.addConsIndicator(
                -_input[i][0] <= -one_dim_center, bin_vars[i], name=name
            )

        return [bin_vars], [output_equal_cons, output_under_half, output_over_half]

    elif outdim == 2:
        # Create additional variables
        max_bin_vars = scip_model.addMatrixVar(
            (n_samples, outdim), vtype="B", name=unique_naming_prefix + "_argmax"
        )

        # Create additional constraints
        output_equal_cons = np.zeros((n_samples, outdim), dtype=object)
        indicator_cons = np.zeros((n_samples, outdim), dtype=object)
        sum_bin_cons = np.zeros((n_samples,), dtype=object)

        # Now populate the constraints
        for i in range(n_samples):
            name = unique_naming_prefix + f"out_eq_{i}_0"
            output_equal_cons[i][0] = scip_model.addCons(
                output[i][0] == max_bin_vars[i][0], name=name
            )
            name = unique_naming_prefix + f"out_eq_{i}_1"
            output_equal_cons[i][1] = scip_model.addCons(
                output[i][1] == max_bin_vars[i][1], name=name
            )
            name = unique_naming_prefix + f"indicator_argmax_{i}_0"
            indicator_cons[i][0] = scip_model.addConsIndicator(
                -_input[i][0] <= -_input[i][1], max_bin_vars[i][0], name=name
            )
            name = unique_naming_prefix + f"indicator_argmax_{i}_1"
            indicator_cons[i][1] = scip_model.addConsIndicator(
                -_input[i][1] <= -_input[i][0], max_bin_vars[i][1], name=name
            )
            name = unique_naming_prefix + f"sum_bin_{i}"
            sum_bin_cons[i] = scip_model.addCons(
                quicksum(max_bin_vars[i][j] for j in range(outdim)) == 1, name=name
            )
        return [max_bin_vars], [output_equal_cons, indicator_cons, sum_bin_cons]
    else:
        # Create additional variables that are needed for classification
        max_bin_vars = scip_model.addMatrixVar(
            (n_samples, outdim), vtype="B", name=unique_naming_prefix + "_argmax"
        )
        slack_vars = scip_model.addMatrixVar(
            (n_samples, outdim), vtype="C", lb=0, name=unique_naming_prefix + "_slack_argmax"
        )
        max_val_vars = scip_model.addMatrixVar(
            (n_samples,), vtype="C", lb=None, ub=None, name=unique_naming_prefix + "_max_val"
        )

        # Create additional constraints that are needed for classification
        output_equal_cons = np.zeros((n_samples, outdim), dtype=object)
        sum_zero_cons = np.zeros((n_samples, outdim), dtype=object)
        sos_slack_bin_cons = np.zeros((n_samples, outdim), dtype=object)
        sum_bin_cons = np.zeros((n_samples,), dtype=object)

        for i in range(n_samples):
            for j in range(outdim):
                name = unique_naming_prefix + f"out_eq_{i}_{j}"
                output_equal_cons[i][j] = scip_model.addCons(
                    output[i][j] == max_bin_vars[i][j], name=name
                )
                name = unique_naming_prefix + f"slack_zero_eq_{i}_{j}"
                sum_zero_cons[i][j] = scip_model.addCons(
                    _input[i][j] + slack_vars[i][j] - max_val_vars[i] == 0, name=name
                )
                name = unique_naming_prefix + f"sos_slack_bin_{i}_{j}"
                sos_slack_bin_cons[i][j] = scip_model.addConsSOS1(
                    [slack_vars[i][j], max_bin_vars[i][j]], name=name
                )

            name = unique_naming_prefix + f"sum_bin_{i}"
            sum_bin_cons[i] = scip_model.addCons(
                quicksum(max_bin_vars[i][j] for j in range(outdim)) == 1, name=name
            )

        return [max_bin_vars, max_val_vars], [
            output_equal_cons,
            sum_zero_cons,
            sos_slack_bin_cons,
            sum_bin_cons,
        ]



"""Module for formulating a
:external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor` or a
:external+sklearn:py:class:`sklearn.tree.DecisionTreeClassifier`
in a PySCIPOpt Model.
"""


from ..modelling import AbstractPredictorConstr
from ..modelling.decision_tree.decision_tree_model import leaf_formulation
from .skgetter import SKgetter


def add_decision_tree_regressor_constr(
    scip_model,
    decision_tree_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs,
):
    """Formulate decision_tree_regressor into a SCIP Model.

    The formulation predicts the values of output_vars using input_vars
    according to decision_tree_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    decision_tree_regressor : :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
        The decision tree regressor to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for decision tree in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for decision tree in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.
    Returns
    -------
    DecisionTreeRegressorConstr
        Object containing information about what was added to scip_model to formulate decision_tree_regressor

    Note
    ----

    |VariablesDimensionsWarn|
    """
    return DecisionTreeConstr(
        scip_model,
        decision_tree_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon,
        False,
        **kwargs,
    )


def add_decision_tree_classifier_constr(
    scip_model,
    decision_tree_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs,
):
    """Formulate decision_tree_classifier into a SCIP Model.

    The formulation predicts the values of output_vars using input_vars
    according to decision_tree_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    decision_tree_classifier : :external+sklearn:py:class:`sklearn.tree.DecisionTreeClassifier`
        The decision tree classifier to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for decision tree in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for decision tree in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.
    Returns
    -------
    DecisionTreeClassifierConstr
        Object containing information about what was added to scip_model to formulate decision_tree_classifier

    Note
    ----

    |VariablesDimensionsWarn|

    Warning
    -------

    Although decision trees with multiple outputs are tested they were never
    used in a non-trivial optimization model. It should be used with care at
    this point.
    """
    return DecisionTreeConstr(
        scip_model,
        decision_tree_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon,
        True,
        **kwargs,
    )


class DecisionTreeConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
    or trained :external+sklearn:py:class:`sklearn.tree.DecisionTreeClassifier` with pyscipopt.

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        epsilon=0.0,
        classification=False,
        formulation="leafs",
        **kwargs,
    ):
        self.epsilon = epsilon
        self.classification = classification
        if self.classification:
            if predictor.n_classes_ <= 2:
                self.output_size = 1
            else:
                self.output_size = predictor.n_classes_
        else:
            self.output_size = predictor.n_outputs_
        SKgetter.__init__(self, predictor, **kwargs)

        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        tree = self.predictor.tree_

        tree_dict = {
            "children_left": tree.children_left,
            "children_right": tree.children_right,
            "feature": tree.feature,
            "threshold": tree.threshold,
            "value": tree.value,
            "node_count": tree.node_count,
            "n_features": tree.n_features,
        }

        new_vars, new_cons = leaf_formulation(
            self.scip_model,
            self.input,
            self.output,
            tree_dict,
            self.unique_naming_prefix,
            self.epsilon,
            classification=self.classification,
        )

        for new_var in new_vars:
            self._created_vars.append(new_var)
        for new_con in new_cons:
            self._created_cons.append(new_con)



""" Utilities for modelling max constraints """

import numpy as np
from pyscipopt import quicksum


def max_formulation(scip_model, _input, output, unique_naming_prefix):
    """
    Create constraints that represent the output of a max function applied to _input.

    The formulation is different depending on the number of features. In the case of there being one:

    Let x be the input \reals, and y the output also in \reals

    .. math::

        \begin{align*}
        y = x
        \end{align*}

    For the case of an arbitrary amount of features, the formulation below is used:

    Let x be the input \reals^{n}, y the output of the max function \reals,
    z be binary variables {0,1}^n,
    and s be the slack variables [0, inf]^{n}

    .. math::

        \begin{align*}
        x_{i} + s_{i} - y == 0 \forall i \in N
        SOS1(s_{i}, z_{i}) \forall i \in N
        \sum z_{i} >= 1
        \end{align*}


    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP Model where the predictor should be inserted.
    _input : np.ndarray
        The (potentially aggregated) output variables from the regression variant of a predictor, which are now
        input to the max formulation.
    output : np.ndarray
        The output variables of the max function
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    Returns
    -------
    created_vars : list
        A list containing np.ndarray of PySCIPOpt variables that were created for the max formulation
    created_cons : list
        A list containing np.ndarray of PySCIPOpt constraints that we created for the max formulation
    """

    assert (output.shape[-1] == 1) and (
        _input.shape[0] == output.shape[0]
    ), f"Input and output dimensions are incorrect. Input: {_input.shape}. Output: {output.shape}"

    # get the in dimensions
    n_samples = _input.shape[0]
    n_features = _input.shape[-1]

    # Separate the formulation into cases
    if n_features == 1:
        name_prefix = unique_naming_prefix + "max"

        # Create additional constraints
        output_equal_cons = np.zeros((n_samples,), dtype=object)

        # Now populate the constraints
        for i in range(n_samples):
            name = unique_naming_prefix + f"out_eq_{i}"
            output_equal_cons[i] = scip_model.addCons(output[i][0] == _input[i][0], name=name)

        return [], [output_equal_cons]
    else:
        # Create additional variables that are needed for the max formulation
        max_bin_vars = scip_model.addMatrixVar(
            (n_samples, n_features), vtype="B", name=unique_naming_prefix + "_bin_max"
        )
        slack_vars = scip_model.addMatrixVar(
            (n_samples, n_features), vtype="C", lb=0, name=unique_naming_prefix + "_slack_max"
        )

        # Create additional constraints that are needed for the max formulation
        max_slack_equal_cons = np.zeros((n_samples, n_features), dtype=object)
        sos_slack_max_cons = np.zeros((n_samples, n_features), dtype=object)
        sum_bin_cons = np.zeros((n_samples,), dtype=object)

        for i in range(n_samples):
            for j in range(n_features):
                name = unique_naming_prefix + f"max_slack_eq_{i}_{j}"
                max_slack_equal_cons[i][j] = scip_model.addCons(
                    output[i][0] == _input[i][j] + slack_vars[i][j], name=name
                )
                name = unique_naming_prefix + f"sos_slack_max_{i}_{j}"
                sos_slack_max_cons[i][j] = scip_model.addConsSOS1(
                    [slack_vars[i][j], max_bin_vars[i][j]], name=name
                )

            name = unique_naming_prefix + f"sum_bin_max_{i}"
            sum_bin_cons[i] = scip_model.addCons(
                quicksum(max_bin_vars[i][j] for j in range(n_features)) >= 1, name=name
            )

        return [max_bin_vars, slack_vars], [max_slack_equal_cons, sos_slack_max_cons, sum_bin_cons]



"""Bases classes for modeling neural network layers."""

import numpy as np

from ...exceptions import ParameterError
from ..base_predictor_constraint import AbstractPredictorConstr
from .activations import (
    add_identity_activation_constraint_layer,
    add_relu_activation_constraint_layer,
    add_sigmoid_activation_constraint_layer,
    add_softmax_activation_constraint_layer,
    add_softplus_activation_constraint_layer,
    add_tanh_activation_constraint_layer,
)


class AbstractNNLayer(AbstractPredictorConstr):
    """Abstract class for NN layers."""

    def __init__(
        self,
        scip_model,
        input_vars,
        output_vars,
        activation,
        unique_naming_prefix,
        **kwargs,
    ):
        self.activation = activation
        if "formulation" in kwargs:
            if kwargs["formulation"] not in ["sos", "bigm"]:
                formulation = kwargs["formulation"]
                raise ParameterError(f"Formulation type {formulation} is neither sos nor bigm")
            self.formulation = kwargs["formulation"]
        else:
            self.formulation = "sos"
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def get_error(self, eps=None):
        # We can't compute externally the error of a layer
        raise AssertionError("Cannot compute the error of an individual layer")

    def _layer_mip_model(self, activation_only=True, **kwargs):
        """Add the layer to model."""
        if self.activation == "relu":
            if self.formulation == "sos":
                slack = self.scip_model.addMatrixVar(
                    (self.input.shape[0], self.output.shape[-1]),
                    vtype="C",
                    lb=0,
                    name=self.unique_naming_prefix + "_slack",
                )
                relu_cons = add_relu_activation_constraint_layer(
                    self, slack, activation_only=activation_only
                )
                self._created_vars.append(slack)
            else:
                activation_vars = self.scip_model.addMatrixVar(
                    (self.input.shape[0], self.output.shape[-1]),
                    vtype="B",
                    name=self.unique_naming_prefix + "_relu_act",
                )
                relu_cons = add_relu_activation_constraint_layer(
                    self, activation_vars, activation_only=activation_only, formulation="bigm"
                )
                self._created_vars.append(activation_vars)
            for cons in relu_cons:
                self._created_cons.append(cons)
        elif self.activation == "logistic" or self.activation == "sigmoid":
            sigmoid_cons = add_sigmoid_activation_constraint_layer(
                self, activation_only=activation_only
            )
            self._created_cons.append(sigmoid_cons)
        elif self.activation == "tanh":
            tanh_cons = add_tanh_activation_constraint_layer(self, activation_only=activation_only)
            self._created_cons.append(tanh_cons)
        elif self.activation == "softmax":
            softmax_cons = add_softmax_activation_constraint_layer(
                self, activation_only=activation_only
            )
            self._created_cons.append(softmax_cons)
        elif self.activation == "softplus":
            softplus_cons = add_softplus_activation_constraint_layer(
                self, activation_only=activation_only
            )
            self._created_cons.append(softplus_cons)
        elif self.activation == "identity" and not activation_only:
            max_bound = self.scip_model.infinity() if self.formulation == "bigm" else 10**5
            affine_cons = add_identity_activation_constraint_layer(self, max_bound)
            self._created_vars.append(affine_cons)
        else:
            if activation_only:
                raise ParameterError(f"Activation layer of type {self.activation} shouldn't exist")
            else:
                raise ParameterError(f"Dense layer of type {self.activation} shouldn't exist")


class ActivationLayer(AbstractNNLayer):
    """Class to build one activation layer of a neural network."""

    def __init__(
        self,
        scip_model,
        output_vars,
        input_vars,
        activation_function,
        unique_naming_prefix,
        **kwargs,
    ):
        super().__init__(
            scip_model,
            input_vars,
            output_vars,
            activation_function,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        self._layer_mip_model(activation_only=True, **kwargs)


class DenseLayer(AbstractNNLayer):
    """Class to build one layer of a neural network."""

    def __init__(
        self,
        scip_model,
        input_vars,
        layer_coefs,
        layer_intercept,
        output_vars,
        activation_function,
        unique_naming_prefix,
        **kwargs,
    ):
        self.coefs = layer_coefs
        self.intercept = layer_intercept
        super().__init__(
            scip_model,
            input_vars,
            output_vars,
            activation_function,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        self.intercept = np.array(self.intercept).reshape(-1)
        self._layer_mip_model(activation_only=False, **kwargs)



"""Module for formulating a
:external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor` or a
:external+sklearn:py:class:`sklearn.ensemble.GradientBoostingClassifier`
into a PySCIPOpt Model.
"""

import numpy as np

from ..modelling import AbstractPredictorConstr
from ..modelling.gradient_boosting import (
    aggregate_estimator_outputs,
    create_aggregation_constraints,
    create_sklearn_tree_estimators,
)
from .skgetter import SKgetter


def add_gradient_boosting_regressor_constr(
    scip_model,
    gradient_boosting_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate gradient_boosting_regressor into scip_model.

    The formulation predicts the values of output_vars using input_vars
    according to gradient_boosting_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    gradient_boosting_regressor : :external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : np.ndarray or list
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : np.ndarray or list, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    GradientBoostingConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    """
    return GradientBoostingConstr(
        scip_model,
        gradient_boosting_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        False,
        **kwargs,
    )


def add_gradient_boosting_classifier_constr(
    scip_model,
    gradient_boosting_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    **kwargs,
):
    """Formulate gradient_boosting_classifier into scip_model.

    The formulation predicts the values of output_vars using input_vars
    according to gradient_boosting_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    gradient_boosting_classifier : :external+sklearn:py:class:`sklearn.ensemble.GradientBoostingClassifier`
        The gradient boosting classifier to insert as predictor.
    input_vars : np.ndarray or list
        Decision variables used as input for gradient boosting classifier in model.
    output_vars : np.ndarray or list, optional
        Decision variables used as output for gradient boosting classifier in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    GradientBoostingConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_classifier.

    Note
    ----
    |VariablesDimensionsWarn|

    """
    return GradientBoostingConstr(
        scip_model,
        gradient_boosting_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        True,
        **kwargs,
    )


class GradientBoostingConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor`
    with SCIP.

    |ClassShort|
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        classification,
        **kwargs,
    ):
        self.estimators_ = []
        self.classification = classification
        SKgetter.__init__(self, predictor, **kwargs)
        if self.classification:
            if predictor.n_classes_ <= 2:
                self.output_size = 1
            else:
                self.output_size = predictor.n_classes_
        else:
            # Gradient Boosting Regression Trees in SKLearn only work for 1-D predictions
            self.output_size = 1
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        if not self.classification:
            assert (
                self.output.shape[-1] == 1
            ), f"Output dimension of SK-learn gradient boosting regressor should be 1. {self.output.shape[-1]} != 1"

        n_samples = self.input.shape[0]
        outdim = self.output.shape[-1]

        # Create individual models for each estimator (decision tree)
        estimators, tree_vars = create_sklearn_tree_estimators(
            self.scip_model,
            self.predictor,
            self.input,
            n_samples,
            outdim,
            self.unique_naming_prefix,
            self.classification,
            gbdt_or_rf="gbdt",
            **kwargs,
        )
        for estimator in estimators:
            self.estimators_.append(estimator)
        self._created_vars.append(tree_vars)

        # Get the constant shift and learning rate used
        lr = self.predictor.learning_rate
        if not self.classification:
            constant = np.array([self.predictor.init_.constant_[0][0] for _ in range(outdim)])
        else:
            constant = self.predictor.init_.predict_log_proba(np.zeros((1, self.input.shape[-1])))[
                0
            ]

        # Add the appropriate constraints for the case of regression and classification
        aggregate_tree_output = aggregate_estimator_outputs(tree_vars, lr, constant, aggr="sum")
        created_vars, created_cons = create_aggregation_constraints(
            self.scip_model,
            aggregate_tree_output,
            self.output,
            n_samples,
            outdim,
            self.unique_naming_prefix,
            self.classification,
        )
        for created_var in created_vars:
            self._created_vars.append(created_var)
        for created_con in created_cons:
            self._created_cons.append(created_con)



"""Module for formulating a :external+sklearn:py:class:`sklearn.cluster.KMeans`
 into a PySCIPOpt Model.
"""

import numpy as np

from ..exceptions import NoModel
from ..modelling import AbstractPredictorConstr
from ..modelling.classification import argmax_bound_formulation
from .skgetter import SKgetter


def add_centroid_cluster_constr(
    scip_model,
    centroid_clusteror,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    formulation="l2",
    **kwargs,
):
    """Formulate centroid_clusteror in scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    centroid_clusteror.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    centroid_clusteror : :external+sklearn:py:class:`sklearn.cluster.KMeans`
        The centroid clusteror to insert as predictor.
    input_vars : list or np.ndarray
        Decision variables used as input for centroid clustering in model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for centroid clustering in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    formulation : str, optional
        The formulation type used when embedding the centroid clustering predictor.
        Valid types are "l2" (standard norm, same as the predictor), and "l1" for a linearised version.
        Warning: The linearised version will incorrectly label some points.

    Returns
    -------
    CentroidClusterConstr
       Object containing information about what was added to scip_model to formulate
       centroid_clusteror.

    Note
    ----
    |VariablesDimensionsWarn|
    """

    return CentroidClusterConstr(
        scip_model,
        centroid_clusteror,
        input_vars,
        output_vars,
        unique_naming_prefix,
        formulation,
        **kwargs,
    )


class CentroidClusterConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained :external+sklearn:py:class:`sklearn.cluster.KMeans`
     with SCIP

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        formulation,
        **kwargs,
    ):
        if hasattr(predictor, "n_clusters"):
            n_clusters = predictor.n_clusters
        else:
            n_clusters = predictor.cluster_centers_.shape[0]
        if n_clusters <= 1:
            raise NoModel(predictor, "Single cluster model is redundant")
        if n_clusters == 2:
            self.output_size = 1
        else:
            self.output_size = n_clusters
        if formulation not in ["l1", "l2"]:
            raise NoModel(predictor, f"Formulation {formulation} is invalid")
        self.formulation = formulation
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the centroids.

        Both X and y should be arrays or lists of variables of conforming dimensions.
        """

        n_samples = self.input.shape[0]
        n_features = self.input.shape[-1]
        if hasattr(self.predictor, "n_clusters"):
            n_clusters = self.predictor.n_clusters
        else:
            n_clusters = self.predictor.cluster_centers_.shape[0]

        # Create additional variables for distance to eac cluster
        dist_vars = self.scip_model.addMatrixVar(
            (n_samples, n_clusters),
            vtype="C",
            lb=None,
            ub=None,
            name=self.unique_naming_prefix + "dist",
        )

        # Create constraints that measure distance from each sample to each centroid
        dist_cons = np.zeros((n_samples, n_clusters), dtype=object)
        if self.formulation == "l1":
            l1_dist_vars = self.scip_model.addMatrixVar(
                (n_samples, n_clusters, n_features, 2),
                vtype="C",
                lb=0,
                ub=None,
                name=self.unique_naming_prefix + "l1_dist",
            )
            l1_dist_cons = np.zeros((n_samples, n_clusters, n_features), dtype=object)
            l1_sos_cons = np.zeros((n_samples, n_clusters, n_features), dtype=object)
        for j in range(n_clusters):
            centroid = self.predictor.cluster_centers_[j]
            for i in range(n_samples):
                if self.formulation == "l1":
                    for k in range(n_features):
                        name = self.unique_naming_prefix + f"l1_dist_{i}_{j}_{k}"
                        l1_dist_cons[i][j][k] = self.scip_model.addCons(
                            self.input[i][k] - centroid[k]
                            == l1_dist_vars[i][j][k][0] - l1_dist_vars[i][j][k][1],
                            name=name,
                        )
                        name = self.unique_naming_prefix + f"l1_sos_{i}_{j}_{k}"
                        l1_sos_cons[i][j][k] = self.scip_model.addConsSOS1(
                            [l1_dist_vars[i][j][k][0], l1_dist_vars[i][j][k][1]], name=name
                        )
                    sum_dist = sum(
                        l1_dist_vars[i][j][k][0] + l1_dist_vars[i][j][k][1]
                        for k in range(n_features)
                    )
                else:
                    sum_dist = sum(
                        (self.input[i][k] - centroid[k]) ** 2 for k in range(n_features)
                    )
                name = self.unique_naming_prefix + f"dist_cons_{i}_{j}"
                dist_cons[i][j] = self.scip_model.addCons(dist_vars[i][j] == sum_dist, name=name)

        # Add created variables and constraints
        if self.formulation == "l1":
            self._created_vars.append(l1_dist_vars)
            self._created_cons.append(l1_dist_cons)
            self._created_cons.append(l1_sos_cons)
        self._created_vars.append(dist_vars)
        self._created_cons.append(dist_cons)

        # Add argmax constraint for closest centroid (turn variables to negative for argmin)
        if n_clusters == 2:
            dist_bin_vars = self.scip_model.addMatrixVar(
                (n_samples, n_clusters), vtype="B", name=self.unique_naming_prefix + "max_dist"
            )
            argmax_vars, argmax_cons = argmax_bound_formulation(
                self.scip_model, -dist_vars, dist_bin_vars, self.unique_naming_prefix
            )
            link_output_cons = np.zeros((n_samples, 1), dtype=object)
            for i in range(n_samples):
                name = self.unique_naming_prefix + f"link_max_out_{i}"
                link_output_cons[i][0] = self.scip_model.addCons(
                    self.output[i][0] == dist_bin_vars[i][1], name=name
                )
            self._created_vars.append(dist_bin_vars)
            self._created_cons.append(link_output_cons)
        else:
            argmax_vars, argmax_cons = argmax_bound_formulation(
                self.scip_model, -dist_vars, self.output, self.unique_naming_prefix
            )
        for new_vars in argmax_vars:
            self._created_vars.append(new_vars)
        for new_cons in argmax_cons:
            self._created_cons.append(new_cons)



"""To get version and git hash of built package."""
from importlib import metadata

try:
    __version__ = metadata.version("pyscipopt_ml")
except metadata.PackageNotFoundError:
    __version__ = "dev"

GIT_HASH = "$Format:%H$"


def get_versions():
    """Get package version."""
    # Downloaded package with inserted git hash.
    if "Format" not in GIT_HASH:
        git_hash = f"-{GIT_HASH}"
    # No inserted git hash, the repo is probably cloned.
    else:
        git_hash = ""

    return {"short": __version__, "long": f"{__version__}{git_hash}"}



"""Module for formulating a :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor` or
:external+sklearn:py:class:`sklearn.neural_network.MLPClassifier` in a PySCIPOpt Model."""

from ..exceptions import NoModel
from ..modelling.classification import argmax_bound_formulation
from ..modelling.neuralnet import BaseNNConstr
from .skgetter import SKgetter


def add_mlp_regressor_constr(
    scip_model, mlp_regressor, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate mlp_regressor into scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    mlp_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    mlp_regressor : :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor`
        The multi-layer perceptron regressor to insert as predictor.
    input_vars : np.ndarray or list
        Decision variables used as input for regression in model.
    output_vars : np.ndarray or list, optional
        Decision variables used as output for regression in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    MLPRegressorConstr
        Object containing information about what was added to scip_model to formulate
        mlp_regressor.

    Raises
    ------
    NoModel
        If the translation to SCIP of the activation function for the network is not
        implemented.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return MLPConstr(
        scip_model, mlp_regressor, input_vars, output_vars, unique_naming_prefix, False, **kwargs
    )


def add_mlp_classifier_constr(
    scip_model, mlp_classifier, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate mlp_classifier into scip_model.

    The formulation predicts the values of output_vars using input_vars according to
    mlp_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The SCIP model where the predictor should be inserted.
    mlp_classifier : :external+sklearn:py:class:`sklearn.neural_network.MLPClassifier`
        The multi-layer perceptron classifier to insert as predictor.
    input_vars : np.ndarray or list
        Decision variables used as input for regression in model.
    output_vars : np.ndarray or list, optional
        Decision variables used as output for regression in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    MLPConstr
        Object containing information about what was added to scip_model to formulate
        mlp_classifier.

    Raises
    ------
    NoModel
        If the translation to SCIP of the activation function for the network is not
        implemented.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return MLPConstr(
        scip_model, mlp_classifier, input_vars, output_vars, unique_naming_prefix, True, **kwargs
    )


class MLPConstr(SKgetter, BaseNNConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor` or
    :external+sklearn:py:class:`sklearn.neural_network.MLPClassifier` with PySCIPOpt.

    |ClassShort|
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        classification=False,
        **kwargs,
    ):
        SKgetter.__init__(self, predictor, **kwargs)
        self.classification = classification
        if self.classification and predictor.n_outputs_ <= 2:
            self.output_size = 1
        else:
            self.output_size = predictor.n_outputs_
        BaseNNConstr.__init__(
            self,
            scip_model,
            predictor,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to SCIP."""
        neural_net = self.predictor
        if neural_net.activation not in self.act_dict:
            raise NoModel(
                neural_net,
                f"No implementation for activation function {neural_net.activation}",
            )
        activation = neural_net.activation

        input_vars = self._input

        for i in range(neural_net.n_layers_ - 1):
            layer_coefs = neural_net.coefs_[i]
            layer_intercept = neural_net.intercepts_[i]

            # For last layer change activation
            if i == neural_net.n_layers_ - 2:
                activation = neural_net.out_activation_
                # Cheat here: In the classification case we don't care about the activation function as the highest
                # value will always have the highest probability and therefore be the selected class
                if self.classification:
                    activation = "identity"
                if self.classification:
                    output = self.scip_model.addMatrixVar(
                        (self.input.shape[0], layer_coefs.shape[-1]),
                        vtype="C",
                        lb=None,
                        ub=None,
                        name=self.unique_naming_prefix + f"layer_{i}",
                    )
                    self._created_vars.append(output)
                else:
                    output = self.output
            else:
                output = self.scip_model.addMatrixVar(
                    (self.input.shape[0], layer_coefs.shape[-1]),
                    vtype="C",
                    lb=None,
                    ub=None,
                    name=self.unique_naming_prefix + f"layer_{i}",
                )
                self._created_vars.append(output)

            layer = self.add_dense_layer(
                input_vars,
                layer_coefs,
                layer_intercept,
                activation,
                output,
                unique_naming_prefix=self.unique_naming_prefix + f"layer_{i}_",
                **kwargs,
            )
            input_vars = layer.output

        if self.classification:
            if neural_net.out_activation_ == "logistic":
                one_dim_center = 0
            else:
                one_dim_center = 0.5
            new_vars, new_cons = argmax_bound_formulation(
                self.scip_model,
                input_vars,
                self.output,
                self.unique_naming_prefix,
                one_dim_center=one_dim_center,
            )
            for added_var in new_vars:
                self._created_vars.append(added_var)
            for added_cons in new_cons:
                self._created_cons.append(added_cons)



"""Module for formulating simple Scikit-Learn Partial Least Squares models in a PySCIPOpt model."""
import numpy as np

from ..modelling import AbstractPredictorConstr
from .skgetter import SKgetter


def add_pls_regression_constr(
    scip_model, pls_regression, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate pls_regression in scip_model.

    The formulation predicts the values of output_vars using input_vars
    according to pls_regression.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The PySCIPOpt model where the predictor will be inserted.
    pls_regression : :external+sklearn:py:class:`sklearn.cross_decomposition.PLSRegression` or
        :external+sklearn:py:class:`sklearn.cross_decomposition.PLSCanonical`
        The partial least squares model to insert.
    input_vars : :list or dict
        Decision variables used as input for partial least squares regression in model.
    output_vars : list or dict
        Decision variables used as output for partial least squares regression in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    PLSRegressionConstr
        Object containing information about what was added to scip_model to
        formulate pls_regression.

    Note
    ----
    |VariablesDimensionsWarn|

    """

    return PLSRegressionConstr(
        scip_model, pls_regression, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


class PLSRegressionConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.cross_decomposition.PLSRegression` or
    :external+sklearn:py:class:`sklearn.cross_decomposition.PLSCanonical` with SCIP

    |ClassShort|.

    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        **kwargs,
    ):
        self.output_size = predictor.n_components
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            scip_model,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def add_regression_constr(self):
        """Add the prediction constraints to SCIP."""

        # Access the PLS model data
        x_mean = self.predictor._x_mean
        x_std = self.predictor._x_std
        coefs = self.predictor.coef_.T
        intercept = self.predictor.intercept_

        # Get the dimensions of the input and output
        n_samples = self.input.shape[0]
        outdim = self.output.shape[-1]

        # Create the PLS model constraints
        pls_cons = np.zeros((n_samples, outdim), dtype=object)
        rhs = (((self.input - x_mean[np.newaxis, :]) / x_std[np.newaxis, :]) @ coefs) + intercept
        for i in range(n_samples):
            for j in range(outdim):
                name = self.unique_naming_prefix + f"pls_reg_{i}_{j}"
                pls_cons[i][j] = self.scip_model.addCons(self.output[i][j] == rhs[i][j], name=name)

        self._created_cons.append(pls_cons)

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to SCIP."""
        self.add_regression_constr()



"""Implements some utility tools for all xgboost objects."""

import json

import numpy as np
from sklearn.base import is_classifier

from ..exceptions import NoModel, NoSolution, ParameterError
from ..modelling import AbstractPredictorConstr


class XGBgetter(AbstractPredictorConstr):
    """Utility class for xgboost models convertors.

    Implement some common functionalities: check predictor is fitted, output dimension, get error

    Attributes
    ----------
    predictor
        Xgboost predictor embedded into SCIP model.
    """

    def __init__(self, predictor, input_vars, output_type="regular", **kwargs):
        if not hasattr(predictor, "_Booster"):
            raise ParameterError(
                "XGBoost model has not yet been fitted. There is nothing to model."
            )
        self.predictor = predictor

    def extract_raw_data_and_create_tree_vars(self, epsilon=0.0):
        """
        Function for extracting information from xgb.Booster and creating additional modelling variables.

        Parameters
        ----------
        epsilon : float, optional
            Small value used to impose strict inequalities for splitting nodes in
            MIP formulations.

        Returns
        -------
        trees : list
            A list of tree dictionaries similar in structure to that provided by SKlearn

        constant : float
            A constant value that is added to all regression output of the decision trees

        tree_vars : np.ndarray
            A numpy array filled with variables that represent output of trees from the trained XGBoost model
        """

        n_samples = self.input.shape[0]
        outdim = self.output.shape[1]

        # Extract the raw data
        xgb_raw = json.loads(self.predictor.get_booster().save_raw(raw_format="json"))
        model_params = xgb_raw["learner"]["learner_model_param"]
        n_trees = len(xgb_raw["learner"]["gradient_booster"]["model"]["trees"])
        constant = float(xgb_raw["learner"]["learner_model_param"]["base_score"])

        # Raise a warning if predictor cannot be modelled
        if int(model_params["num_class"]) > 1 and int(model_params["num_target"]) > 1:
            raise NoModel(self.predictor, "Multi-regression for multiple classes is not allowed")
        booster_type = xgb_raw["learner"]["gradient_booster"]["name"]
        if booster_type != "gbtree":
            raise NoModel(self.predictor, f"model not implemented for {booster_type}")
        assert (
            int(xgb_raw["learner"]["learner_model_param"]["num_target"]) == outdim
            or int(xgb_raw["learner"]["learner_model_param"]["num_class"]) == outdim
        ), "Dimension error when constructing XGBoost representation."
        assert (
            n_trees % outdim == 0
        ), "Uneven amount of estimator trees per class / output for XGBoost.Booster"

        # Convert the raw data into a format similar to sk-learn
        trees = [[{} for _ in range(outdim)] for _ in range(self.predictor.n_estimators)]
        trees_converted = [0 for _ in range(outdim)]
        for i in range(n_trees):
            tree_class = xgb_raw["learner"]["gradient_booster"]["model"]["tree_info"][i]
            tree_raw = xgb_raw["learner"]["gradient_booster"]["model"]["trees"][i]
            tree = trees[trees_converted[tree_class]][tree_class]
            tree["threshold"] = np.array(tree_raw["split_conditions"], dtype=np.float32)
            tree["children_left"] = np.array(tree_raw["left_children"])
            tree["children_right"] = np.array(tree_raw["right_children"])
            tree["feature"] = np.array(tree_raw["split_indices"])
            tree["node_count"] = len(tree_raw["split_conditions"])  # Also accessible via num_nodes
            tree["n_features"] = int(tree_raw["tree_param"]["num_feature"])
            tree["value"] = tree["threshold"].reshape((tree["node_count"], 1, 1))
            trees_converted[tree_class] += 1
        for i in range(outdim):
            if trees_converted[i] != trees_converted[0]:
                raise ParameterError("Uneven amount of estimators per output")

        shape = (n_samples, n_trees // outdim, outdim)
        tree_vars = self.scip_model.addMatrixVar(shape, vtype="C", lb=None, ub=None, name="tree")

        return trees, constant, tree_vars

    def get_error(self, eps=None):
        """
        Returns error in SCIP's solution with respect to the actual output of the trained predictor

        Parameters
        ----------
        eps : float or int or None, optional
            The maximum allowed tolerance for a mismatch between the actual predictive model and SCIP.
            If the error is larger than eps an appropriate warning is printed

        Returns
        -------
        error: np.ndarray
            The absolute values of the difference between SCIP's solution and the trained ML model's output given
            the input as defined by SCIP. The matrix is the same dimension as the output of the trained predictor.
            Using sklearn / pyscipopt, the absolute difference between model.predict(input) and scip.getVal(output).

        Raises
        ------
        NoSolution
            If SCIP has no solution (either was not optimized or is infeasible).
        """

        if self._has_solution:
            if not is_classifier(self.predictor):
                xgb_output_values = self.predictor.predict(self.input_values).reshape(
                    self.input.shape[0], self.output.shape[-1]
                )
            else:
                xgb_class_prediction = self.predictor.predict(self.input_values)
                xgb_output_values = np.zeros((self.input.shape[0], self.output.shape[-1]))
                for i, class_pred in enumerate(xgb_class_prediction):
                    if self.output.shape[-1] == 1:
                        xgb_output_values[i][0] = class_pred
                    else:
                        xgb_output_values[i][class_pred] = 1
            scip_output_values = self.output_values
            error = np.abs(xgb_output_values - scip_output_values)
            max_error = np.max(error)
            if eps is not None and max_error > eps:
                print(
                    f"SCIP output values of ML model {self.predictor} have larger than max error {max_error} > {eps}"
                )
            return error

        raise NoSolution()



"""Module for formulating a XGBoost gradient boosting or random forest regressor / classifier
 into a PySCIPOpt Model. """

import numpy as np

from ..modelling import AbstractPredictorConstr
from ..modelling.gradient_boosting import aggregated_estimator_formulation
from .xgbgetter import XGBgetter


def add_xgbregressor_constr(
    scip_model,
    xgboost_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate xgboost_regressor (gradient boosting decision tree) as constraints into a pyscipopt Model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    xgboost_regressor : :external+xgb:py:class:`xgboost.XGBRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    XGBoostRegressorConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return XGBoostConstr(
        scip_model,
        xgboost_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        aggr="sum",
        classification=False,
        **kwargs,
    )


def add_xgbclassifier_constr(
    scip_model,
    xgboost_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate xgboost_classifier (gradient boosting decision tree) as constraints into a pyscipopt Model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    xgboost_classifier : :external+xgb:py:class:`xgboost.XGBClassifier`
        The gradient boosting classifier to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    XGBoostClassifierConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_classifier.

    Note
    ----
    |VariablesDimensionsWarn|

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """

    return XGBoostConstr(
        scip_model,
        xgboost_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        aggr="sum",
        classification=True,
        **kwargs,
    )


def add_xgbregressor_rf_constr(
    scip_model,
    xgboost_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate xgboost_regressor (random forest) as constraints into a pyscipopt Model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    xgboost_regressor : :external+xgb:py:class:`xgboost.XGBRFRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    XGBoostRegressorConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """
    return XGBoostConstr(
        scip_model,
        xgboost_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        aggr="sum",
        classification=False,
        **kwargs,
    )


def add_xgbclassifier_rf_constr(
    scip_model,
    xgboost_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate xgboost_classifier (random forest) as constraints into a pyscipopt Model.

    The formulation predicts the values of output_vars using input_vars
    according to xgboost_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    xgboost_classifier : :external+xgb:py:class:`xgboost.XGBClassifier`
        The gradient boosting classifier to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    XGBoostClassifierConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_classifier.

    Note
    ----
    |VariablesDimensionsWarn|

    Raises
    ------
    NoModel
        If the booster is not of type "gbtree".
    """

    return XGBoostConstr(
        scip_model,
        xgboost_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        aggr="sum",
        classification=True,
        **kwargs,
    )


class XGBoostConstr(XGBgetter, AbstractPredictorConstr):
    """Class to model trained :external+xgb:py:class:`xgboost.XGBRegressor` or
    :external+xgb:py:class:`xgboost.XGBClassifier` with constraints in a pyscipopt Model.

    |ClassShort|
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix="",
        epsilon=0.0,
        aggr="sum",
        classification=False,
        **kwargs
    ):
        XGBgetter.__init__(self, predictor, input_vars, **kwargs)
        self.estimators_ = []
        self.classification = classification
        self.epsilon = epsilon
        self.aggr = aggr
        # Get the output dimension
        if hasattr(predictor, "predict_proba"):
            dummy_prediction = predictor.predict_proba(np.zeros((1, predictor.n_features_in_)))
            output_size = dummy_prediction.shape[-1]
            if self.classification and output_size <= 2:
                self.output_size = 1
            else:
                self.output_size = output_size
        else:
            dummy_prediction = predictor.predict(np.zeros((1, predictor.n_features_in_)))
            self.output_size = dummy_prediction.shape[-1]
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):

        # Extract the information on the trained model and create new variables
        trees, constant, tree_vars = self.extract_raw_data_and_create_tree_vars(
            epsilon=self.epsilon
        )
        self._created_vars.append(tree_vars)

        # Construct each individual tree estimator and the constraints that link them
        constant = np.array([constant for _ in range(self.output.shape[-1])])
        estimators, created_vars, created_cons = aggregated_estimator_formulation(
            self.scip_model,
            self.input,
            self.output,
            tree_vars,
            trees,
            constant,
            1,
            self.predictor.n_estimators,
            self.unique_naming_prefix,
            self.epsilon,
            self.aggr,
            self.classification,
            **kwargs,
        )

        for estimator in estimators:
            self.estimators_.append(estimator)
        for created_var in created_vars:
            self._created_vars.append(created_var)
        for created_con in created_cons:
            self._created_cons.append(created_con)



"""Module for inserting simple Scikit-Learn regression models into a PySCIPOpt model.

All linear models should work:
   - :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
   - :external+sklearn:py:class:`sklearn.linear_model.Ridge`
   - :external+sklearn:py:class:`sklearn.linear_model.Lasso`
   - :external+sklearn:py:class:`sklearn.linear_model.ElasticNet`
"""
import numpy as np

from ..modelling import AbstractPredictorConstr
from .skgetter import SKgetter


class BaseSKlearnRegressionConstr(SKgetter, AbstractPredictorConstr):
    """Predict output variables using Linear Regression that takes PySCIPOpt variables as input.

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        **kwargs,
    ):
        SKgetter.__init__(self, predictor, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            scip_model,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def add_regression_constr(self, output=None):
        """Add the prediction constraints to SCIP."""

        if output is None:
            output = self.output

        coefs = self.predictor.coef_.T
        intercept = self.predictor.intercept_

        # Transform the input by an affine transformation (Ax + b)
        affine = self.input @ coefs + intercept
        if affine.shape != output.shape:
            affine = affine.reshape(output.shape)
        affine_cons = np.zeros(affine.shape, dtype=object)

        # Iterate through all entries and create the appropriate constraints
        for i in range(output.shape[0]):
            for j in range(output.shape[1]):
                name = self.unique_naming_prefix + f"linreg_{i}_{j}"
                affine_cons[i][j] = self.scip_model.addCons(
                    affine[i][j] == output[i][j], name=name
                )

        self._created_cons.append(affine_cons)



"""Module for formulating a LightGBM gradient boosting or random forest regressor / classifier into
a PySCIPOpt Model."""

import numpy as np

from ..modelling import AbstractPredictorConstr
from ..modelling.gradient_boosting import aggregated_estimator_formulation
from .lgbgetter import LGBgetter


def add_lgbregressor_constr(
    scip_model,
    lightgbm_regressor,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate lightgbm_regressor as constraints into a pyscipopt Model. Accommodates both Gradient Boosting
    Decision Trees and Random Forests as boosting types.

    The formulation predicts the values of output_vars using input_vars
    according to lightgbm_regressor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    lightgbm_regressor : :external+lgb:py:class:`lightgbm.LGBMRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting regressor in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    LightGBMRegressorConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_regressor.

    Raises
    ------
    NoModel
        If the boosting type is not "gbdt" or "rf".

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return LightGBMConstr(
        scip_model,
        lightgbm_regressor,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        classification=False,
        **kwargs,
    )


def add_lgbclassifier_constr(
    scip_model,
    lightgbm_classifier,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    epsilon=0.0,
    **kwargs
):
    """Formulate lightgbm_classifier as constraints into a pyscipopt Model. Accommodates both Gradient Boosting
    Decision Trees and Random Forests as boosting types.

    The formulation predicts the values of output_vars using input_vars
    according to lightgbm_classifier.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt Model where the predictor should be inserted.
    lightgbm_classifier : :external+lgb:py:class:`lightgbm.LGBMClassifier`
        The gradient boosting classifier to insert as predictor.
    input_vars : list or dict
        Decision variables used as input for gradient boosting classifier in model.
    output_vars : list or dict, optional
        Decision variables used as output for gradient boosting classifier in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.

    Returns
    -------
    LightGBMClassifierConstr
        Object containing information about what was added to scip_model to formulate
        gradient_boosting_classifier.

    Raises
    ------
    NoModel
        If the boosting type is not "gbdt" or "rf".

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return LightGBMConstr(
        scip_model,
        lightgbm_classifier,
        input_vars,
        output_vars,
        unique_naming_prefix,
        epsilon=epsilon,
        classification=True,
        **kwargs,
    )


class LightGBMConstr(LGBgetter, AbstractPredictorConstr):
    """Class to model trained :external+lgb:py:class:`lightgbm.LGBMRegressor` or
     :external+lgb:py:class:`lightgbm.LGBMClassifier` with constraints in a pyscipopt Model.

    |ClassShort|
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars,
        unique_naming_prefix="",
        epsilon=0.0,
        classification=False,
        **kwargs
    ):
        LGBgetter.__init__(self, predictor, input_vars, **kwargs)
        self.estimators_ = []
        self.classification = classification
        self.epsilon = epsilon
        if self.classification:
            if predictor.n_classes_ <= 2:
                self.output_size = 1
            else:
                self.output_size = predictor.n_classes_
        else:
            # LGB only accommodates single output regression
            self.output_size = 1
        AbstractPredictorConstr.__init__(
            self, scip_model, input_vars, output_vars, unique_naming_prefix, **kwargs
        )

    def _mip_model(self, **kwargs):

        # Extract the information on the trained model and create new variables
        trees, tree_vars = self.extract_raw_data_and_create_tree_vars(epsilon=self.epsilon)
        self._created_vars.append(tree_vars)

        # Construct each individual tree estimator and the constraints that link them
        constant = np.zeros((self.output.shape[-1],))
        aggr = "sum" if self.predictor.boosting_type == "gbdt" else "avg"
        estimators, created_vars, created_cons = aggregated_estimator_formulation(
            self.scip_model,
            self.input,
            self.output,
            tree_vars,
            trees,
            constant,
            1,
            self.predictor.n_estimators,
            self.unique_naming_prefix,
            self.epsilon,
            aggr,
            self.classification,
            **kwargs,
        )

        for estimator in estimators:
            self.estimators_.append(estimator)
        for created_var in created_vars:
            self._created_vars.append(created_var)
        for created_con in created_cons:
            self._created_cons.append(created_con)



from .centroid_cluster import add_centroid_cluster_constr
from .decision_tree import (
    add_decision_tree_classifier_constr,
    add_decision_tree_regressor_constr,
)
from .gradient_boosting import (
    add_gradient_boosting_classifier_constr,
    add_gradient_boosting_regressor_constr,
)
from .linear_regression import add_linear_regression_constr
from .logistic_regression import add_logistic_regression_constr
from .mlp import add_mlp_classifier_constr, add_mlp_regressor_constr
from .multi_output import (
    add_multi_output_classifier_constr,
    add_multi_output_regressor_constr,
)
from .pipeline import add_pipeline_constr
from .pls import add_pls_regression_constr
from .random_forest import (
    add_random_forest_classifier_constr,
    add_random_forest_regressor_constr,
)
from .support_vector import (
    add_support_vector_classifier_constr,
    add_support_vector_regressor_constr,
)



"""Module for formulating a
:external+sklearn:py:class:`sklearn.linear_model.LogisticRegression` in a
PySCIPOpt Model.
"""

import numpy as np
from pyscipopt import exp, quicksum

from ..exceptions import NoModel, ParameterError
from ..modelling.classification import argmax_bound_formulation
from .base_regression import BaseSKlearnRegressionConstr


def add_logistic_regression_constr(
    scip_model,
    logistic_regression,
    input_vars,
    output_vars=None,
    unique_naming_prefix="",
    output_type="classification",
    **kwargs,
):
    """Formulate logistic_regression in a SCIP Model.

    The formulation predicts the values of output_vars using input_vars according to
    logistic_regression.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The PySCIPOpt model where the predictor will be inserted.
    logistic_regression : :external+sklearn:py:class:`sklearn.linear_model.LogisticRegression`
        The logistic regression model to insert.
    input_vars : :list or dict
        Decision variables used as input for logistic regression in model.
    output_vars : list or dict
        Decision variables used as output for logistic regression in model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    output_type : {'classification', 'regression'}, default='classification'
        If the option chosen is 'classification' the output is 1 for exactly one class and
        0 for all others. If the option
        'regression' is chosen the output is the probability of each class.

    Returns
    -------
    LogisticRegressionConstr
        Object containing information about what was added to scip_model to formulate
        logistic_regression.

    Raises
    ------

    NoModel
        If the logistic regression is not a binary label regression

    ParameterError
        If the value of output_type is set to a non-conforming value (see above).

    Warning
    -------

    When there is a (near) tie for the most likely class, users should be aware that SCIP tolerances will allow
    either of the two classes to be selected. In a scenario where there are two classes, and the logistic regression
    model assigns probability of class 1 as 0.5000000001. Naturally this is larger than 0.4999999999 assigned to
    class 2, however from a numerics point of view in SCIP both can be selected.

    Note
    ----
    |VariablesDimensionsWarn|

    """
    return LogisticRegressionConstr(
        scip_model,
        logistic_regression,
        input_vars,
        output_vars,
        unique_naming_prefix,
        output_type,
        **kwargs,
    )


class LogisticRegressionConstr(BaseSKlearnRegressionConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.linear_model.LogisticRegression` with SCIP

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        output_type="classification",
        **kwargs,
    ):
        if output_type not in ("classification", "regression"):
            raise ParameterError("output_type should be either 'classification' or 'regression'")

        self.output_type = output_type
        self.output_size = 1 if predictor.classes_.size <= 2 else predictor.classes_.size

        BaseSKlearnRegressionConstr.__init__(
            self,
            scip_model,
            predictor,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to the SCIP Model."""

        if self.input.ndim > 2 or self.output.ndim > 2:
            raise NoModel(
                self.predictor,
                "Logistic regression only supported for input.ndim <=2 and output.ndim <=2."
                f"Dim: {self.input.ndim} / {self.output.ndim}",
            )

        n_samples = self.input.shape[0]
        outdim = self.output.shape[-1]

        # First model the linear transformation
        affine_vars = self.scip_model.addMatrixVar(
            self.output.shape,
            vtype="C",
            lb=None,
            ub=None,
            name=self.unique_naming_prefix + "affine",
        )
        self._created_vars.append(affine_vars)
        self.add_regression_constr(output=affine_vars)

        # In the case of classification, we need to return binary results.
        if self.output_type == "classification":
            # Use an argmax formulation on the output of the affine transformation
            added_vars, added_cons = argmax_bound_formulation(
                self.scip_model,
                affine_vars,
                self.output,
                self.unique_naming_prefix,
                one_dim_center=0,
            )

            # Store all the appropriate variables
            for added_var in added_vars:
                self._created_vars.append(added_var)
            for added_con in added_cons:
                self._created_cons.append(added_con)
        else:
            # In this case we are turning probabilities and need to model the logistic function explicitly
            if outdim == 1:
                # In this case we are dealing with only a single class
                log_output_vars = self.scip_model.addMatrixVar(
                    self.output.shape,
                    vtype="C",
                    lb=0,
                    ub=1,
                    name=self.unique_naming_prefix + "logistic",
                )
                logistic_cons = self.scip_model.addMatrixCons(
                    log_output_vars == 1 / (1 + exp(-affine_vars)),
                    name=self.unique_naming_prefix + "sigmoid",
                )
                output_eq_cons = self.scip_model.addMatrixCons(
                    log_output_vars == self.output, name=self.unique_naming_prefix + "output_eq"
                )
                self._created_vars.append(log_output_vars)
                self._created_cons.append(logistic_cons)
                self._created_cons.append(output_eq_cons)
            else:
                log_output_vars = self.scip_model.addMatrixVar(
                    self.output.shape,
                    vtype="C",
                    lb=0,
                    ub=1,
                    name=self.unique_naming_prefix + "logistic",
                )
                # Get the actual predictive model used by sklearn in the logistic regression model
                ovr = self.predictor.multi_class in ["ovr", "warn"] or (
                    self.predictor.multi_class == "auto"
                    and (
                        self.predictor.classes_.size <= 2
                        or self.predictor.solver in ("liblinear", "newton-cholesky")
                    )
                )

                # If the model is one-vs-rest then model using logistic (sigmoid). Otherwise we use softmax
                cons_probability = np.zeros(self.output.shape, dtype=object)
                output_eq_cons = np.zeros(self.output.shape, dtype=object)
                for i in range(n_samples):
                    sum_logistic = quicksum(
                        1 / (1 + exp(-affine_vars[i][j])) for j in range(affine_vars[i].shape[0])
                    )
                    sum_softmax = quicksum(
                        exp(affine_vars[i][j]) for j in range(affine_vars[i].shape[0])
                    )
                    for j in range(outdim):
                        name = f"logistic_{i}_{j}"
                        if ovr:
                            logistic = 1 / (1 + exp(-affine_vars[i][j]))
                            cons_probability[i][j] = self.scip_model.addCons(
                                logistic / sum_logistic == log_output_vars[i][j], name=name
                            )
                        else:
                            cons_probability[i][j] = self.scip_model.addCons(
                                exp(affine_vars[i][j]) / sum_softmax == log_output_vars[i][j],
                                name=name,
                            )
                        name = self.unique_naming_prefix + f"output_eq_{i}_{j}"
                        output_eq_cons[i][j] = self.scip_model.addCons(
                            log_output_vars[i][j] == self.output[i][j], name=name
                        )
                self._created_vars.append(log_output_vars)
                self._created_cons.append(cons_probability)
                self._created_cons.append(output_eq_cons)



import sys


def sklearn_convertors():
    """Collect convertors for Scikit-learn objects."""
    if "sklearn" in sys.modules:
        from .sklearn import (
            add_centroid_cluster_constr,
            add_decision_tree_classifier_constr,
            add_decision_tree_regressor_constr,
            add_gradient_boosting_classifier_constr,
            add_gradient_boosting_regressor_constr,
            add_linear_regression_constr,
            add_logistic_regression_constr,
            add_mlp_classifier_constr,
            add_mlp_regressor_constr,
            add_multi_output_classifier_constr,
            add_multi_output_regressor_constr,
            add_pipeline_constr,
            add_pls_regression_constr,
            add_random_forest_classifier_constr,
            add_random_forest_regressor_constr,
            add_support_vector_classifier_constr,
            add_support_vector_regressor_constr,
        )

        return {
            "LinearRegression": add_linear_regression_constr,
            "Ridge": add_linear_regression_constr,
            "Lasso": add_linear_regression_constr,
            "ElasticNet": add_linear_regression_constr,
            "LogisticRegression": add_logistic_regression_constr,
            "DecisionTreeClassifier": add_decision_tree_classifier_constr,
            "DecisionTreeRegressor": add_decision_tree_regressor_constr,
            "GradientBoostingRegressor": add_gradient_boosting_regressor_constr,
            "GradientBoostingClassifier": add_gradient_boosting_classifier_constr,
            "RandomForestRegressor": add_random_forest_regressor_constr,
            "RandomForestClassifier": add_random_forest_classifier_constr,
            "MLPRegressor": add_mlp_regressor_constr,
            "MLPClassifier": add_mlp_classifier_constr,
            "PLSRegression": add_pls_regression_constr,
            "PLSCanonical": add_pls_regression_constr,
            "LinearSVR": add_support_vector_regressor_constr,
            "SVR": add_support_vector_regressor_constr,
            "LinearSVC": add_support_vector_classifier_constr,
            "SVC": add_support_vector_classifier_constr,
            "KMeans": add_centroid_cluster_constr,
            "MiniBatchKMeans": add_centroid_cluster_constr,
            "Pipeline": add_pipeline_constr,
            "MultiOutputClassifier": add_multi_output_classifier_constr,
            "MultiOutputRegressor": add_multi_output_regressor_constr,
        }

    return {}


def pytorch_convertors():
    """Collect known PyTorch objects that can be formulated and the conversion class."""
    if "torch" in sys.modules:
        import torch  # pylint: disable=import-outside-toplevel

        from .torch import (  # pylint: disable=import-outside-toplevel
            add_sequential_constr,
        )

        return {torch.nn.Sequential: add_sequential_constr}
    return {}


def xgboost_convertors():
    """Collect known XGBoost objects that can be formulated and the conversion class."""
    if "xgboost" in sys.modules:
        import xgboost as xgb  # pylint: disable=import-outside-toplevel

        from .xgboost import (  # pylint: disable=import-outside-toplevel
            add_xgbclassifier_constr,
            add_xgbclassifier_rf_constr,
            add_xgbregressor_constr,
            add_xgbregressor_rf_constr,
        )

        return {
            xgb.XGBRegressor: add_xgbregressor_constr,
            xgb.XGBClassifier: add_xgbclassifier_constr,
            xgb.XGBRFRegressor: add_xgbregressor_rf_constr,
            xgb.XGBRFClassifier: add_xgbclassifier_rf_constr,
        }
    return {}


def lightgbm_convertors():
    """Collect known LightGBM objects that can be formulated and the conversion class."""
    if "lightgbm" in sys.modules:
        import lightgbm as lgb  # pylint: disable=import-outside-toplevel

        from .lightgbm import (  # pylint: disable=import-outside-toplevel
            add_lgbclassifier_constr,
            add_lgbregressor_constr,
        )

        return {
            lgb.LGBMRegressor: add_lgbregressor_constr,
            lgb.LGBMClassifier: add_lgbclassifier_constr,
        }
    return {}


def keras_convertors():
    """Collect known Keras objects that can be formulated and the conversion class."""
    if "tensorflow" in sys.modules:
        from tensorflow import keras  # pylint: disable=import-outside-toplevel

        from .keras import add_keras_constr  # pylint: disable=import-outside-toplevel

        return {
            keras.Sequential: add_keras_constr,
            keras.Model: add_keras_constr,
        }
    return {}


def onnx_convertors():
    """Collect known ONNX objects that can be formulated and the conversion class."""
    if "onnx" in sys.modules:
        import onnx  # pylint: disable=import-outside-toplevel

        from .onnx import add_onnx_constr  # pylint: disable=import-outside-toplevel

        return {
            onnx.ModelProto: add_onnx_constr,
            onnx.onnx_ml_pb2.ModelProto: add_onnx_constr,
        }

    return {}


def registered_predictors():
    """Return the list of registered predictors."""
    convertors = {
        **sklearn_convertors(),
        **pytorch_convertors(),
        **xgboost_convertors(),
        **lightgbm_convertors(),
        **keras_convertors(),
        **onnx_convertors(),
    }
    return convertors



import numpy as np
from pyscipopt import Model
from sklearn.neural_network import MLPClassifier
from utils import read_csv_to_dict

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of a conglomeration of airlines.
This set of airlines is worried about current customer satisfaction levels,
and has assigned a budget to improve the predicted levels. Each airline, however,
has its own customer base, and incentives for which bit of the flight experience to
improve. Using the budget, maximise the total improved customer satisfaction levels.
"""

# Path to water potability data
data_dict = read_csv_to_dict("./tests/data/airline_satisfaction.csv")

# The features of our predictor. All distance based features are variables.
features = [
    ("Gender", "Male"),
    ("Customer Type", "Loyal"),
    ("Type of Travel", "Business travel"),
    ("Class", "Eco"),
    ("Flight Distance", int),
    ("Inflight wifi service", int),
    ("Departure/Arrival time convenient", int),
    ("Ease of Online booking", int),
    ("Gate location", int),
    ("Food and drink", int),
    ("Online boarding", int),
    ("Seat comfort", int),
    ("Inflight entertainment", int),
    ("On-board service", int),
    ("Leg room service", int),
    ("Baggage handling", int),
    ("Checkin service", int),
    ("Inflight service", int),
    ("Cleanliness", int),
]
feature_to_idx = {feature: i for i, feature in enumerate(features)}
n_features = len(features)
n_passengers = 200
np.random.seed(42)

# Generate the actual input data arrays for the ML predictors
X = []
y = np.array([x == "satisfied" for x in data_dict["satisfaction"]]).astype(int).reshape(-1, 1)
for feature, feature_key in features:
    if feature == "Class":
        X.append(np.array([("Eco" not in x) for x in data_dict[feature]]).astype(int))
    elif feature_key != int:
        X.append(np.array([x == feature_key for x in data_dict[feature]]).astype(int))
    else:
        X.append(np.array([int(x) for x in data_dict[feature]]))
X = np.swapaxes(np.array(X), 0, 1)



import numpy as np
from pyscipopt import Model
from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from utils import read_csv_to_dict, train_torch_neural_network

from src.pyscipopt_ml.add_predictor import add_predictor_constr

"""
In this scenario we take the point of view of an auto manufacturer.

We have access to open source data (thanks to: https://www.kaggle.com/datasets/gagandeep16/car-sales),
for which we can build predictors. The predictors takes as input a variety of features for the
given vehicle, and output either the price, the resell value, or the number of units that will sell.
These properties are dependent on each other, so the output of one predictor can also be
input into another predictor.

The goal of this MIP is to design a vehicle that will sell the maximum amount of units possible,
while being sufficiently different from any of the other most popular vehicles.
Additional constraints are added to ensure that the designed vehicle has a high resell value,
and is fuel efficient.

Let x be the vector of static features of the designed vehicle, i.e., those that cannot be changed.
Price, resell value, and amount sold for instance are not entirely controllable, but engine size is.

Let f be the ML predictor that predicts the amount sold of the designed vehicle.
Let g be the ML predictor that predicts the price of the designed vehicle.
Let h be the ML predictor that predicts the resell value of the designed vehicle.

Let V be the index set of best selling vehicles that the designed vehicle must be sufficiently different to.
Let I be the index set of features
Let auto_val[v][i] be the value of feature i for best selling vehicle v
Let e be some minimum level of fuel efficiency
Let c be some minimum ratio of resale value compared to the actual fresh purhcase price

To ensure sufficiently different we use look at the relative difference of each feature.
So for a given vehicle v, and feature i, we know that it has value 5. We also know that feature i must
be in the range [2, 10]. So if our designed vehicle has value 6, then the difference is abs(5-6) / (10-2)

x[i] - auto_val[v][i] == pos_slack[v][i] - neg_slack[v][i] for all V, I
pos_slack[v][i] <= (x[i].UB - x[i].LB) * bin_slack[v][i] for all V, I
neg_slack[v][i] <= (x[i].UB - x[i].LB) * (1 - bin_slack[v][i]) for all V, I
sum_diff[v] = sum_i (pos_slack[v][i] + neg_slack[v][i]) / (x[i].UB - x[i].LB) for all V
sum_diff[v] >= 2 for all V

x[i] >= e where i is the fuel efficiency index

price = g(x)
amount_sold = f(x, price)
resale = h(x, price)

resale >= c * price

max(amount_sold)
"""


def build_and_optimise_auto_manufacturer(
    training_seed=42,
    data_seed=42,
    gbdt_rf_or_mlp="gbdt",
    formulation="sos",
    framework="sklearn",
    max_depth_or_layer_size=6,
    n_estimators_or_layers=6,
    build_only=False,
):
    # Initialise a numpy random state
    data_random_state = np.random.RandomState(data_seed)
    training_random_state = np.random.RandomState(training_seed)

    # Path to car data
    data_dict = read_csv_to_dict("./tests/data/car_sales.csv")

    # Get an array of the features
    features = [
        "Vehicle_type",
        "Engine_size",
        "Horsepower",
        "Wheelbase",
        "Width",
        "Length",
        "Curb_weight",
        "Fuel_capacity",
        "Fuel_efficiency",
        "Power_perf_factor",
    ]
    n_features = len(features)
    n_car_comparisons = 20

    amount_sold_key = "Sales_in_thousands"
    price_key = "Price_in_thousands"
    resale_value_key = "__year_resale_value"

    # Generate the actual input data arrays for the ML predictors
    X = []
    amount_sales = np.array([float(x) for x in data_dict[amount_sold_key]]).reshape(-1, 1)
    price = np.array([float(x) for x in data_dict[price_key]]).reshape(-1, 1)
    resale_price = np.array([float(x) for x in data_dict[resale_value_key]]).reshape(-1, 1)
    for feature in features:
        if feature == "Vehicle_type":
            X.append((np.array(data_dict[feature]) == "Passenger").astype(int))
        else:
            X.append(np.array([float(x) for x in data_dict[feature]]))
    X = np.swapaxes(np.array(X), 0, 1)

    # Create the prediction models
    if gbdt_rf_or_mlp == "gbdt":
        reg_sales = GradientBoostingRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(np.concatenate((X, price), axis=1), amount_sales.reshape(-1))
        reg_price = GradientBoostingRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(X, price.reshape(-1))
        reg_resale = GradientBoostingRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(np.concatenate((X, price), axis=1), resale_price.reshape(-1))
    elif gbdt_rf_or_mlp == "rf":
        reg_sales = RandomForestRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(np.concatenate((X, price), axis=1), amount_sales.reshape(-1))
        reg_price = RandomForestRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(X, price.reshape(-1))
        reg_resale = RandomForestRegressor(
            n_estimators=n_estimators_or_layers,
            max_depth=max_depth_or_layer_size,
            random_state=training_random_state,
        ).fit(np.concatenate((X, price), axis=1), resale_price.reshape(-1))
    elif gbdt_rf_or_mlp == "mlp":
        if framework == "sklearn":
            hidden_layer_sizes = tuple(
                [max_depth_or_layer_size for i in range(n_estimators_or_layers)]
            )
            reg_sales = MLPRegressor(
                random_state=training_random_state, hidden_layer_sizes=hidden_layer_sizes
            ).fit(np.concatenate((X, price), axis=1), amount_sales.reshape(-1))
            reg_price = MLPRegressor(
                random_state=training_random_state, hidden_layer_sizes=hidden_layer_sizes
            ).fit(X, price.reshape(-1))
            reg_resale = MLPRegressor(
                random_state=training_random_state, hidden_layer_sizes=hidden_layer_sizes
            ).fit(np.concatenate((X, price), axis=1), resale_price.reshape(-1))
        else:
            reg_price = train_torch_neural_network(
                X,
                price,
                n_estimators_or_layers,
                max_depth_or_layer_size,
                training_seed,
                reshape=False,
            )
            reg_sales = train_torch_neural_network(
                np.concatenate((X, price), axis=1),
                amount_sales,
                n_estimators_or_layers,
                max_depth_or_layer_size,
                training_seed,
                reshape=False,
            )
            reg_resale = train_torch_neural_network(
                np.concatenate((X, price), axis=1),
                resale_price,
                n_estimators_or_layers,
                max_depth_or_layer_size,
                training_seed,
                reshape=False,
            )
    else:
        raise ValueError(f"Unknown value: {gbdt_rf_or_mlp}")

    # Create the SCIP Model
    scip = Model()

    # Create variables deciding the features of the manufactured car and the predicted sales information
    feature_vars = np.zeros((1, n_features), dtype=object)
    amount_sold_vars = np.zeros((1, 1), dtype=object)
    price_vars = np.zeros((1, 1), dtype=object)
    resale_vars = np.zeros((1, 1), dtype=object)

    for i in range(n_features):
        if i == 0:
            feature_vars[0][i] = scip.addVar(vtype="B", name=f"feature_{i}")
        else:
            max_val = np.max(X[:, i])
            min_val = np.min(X[:, i])
            lb = max(0, min_val - 0.1 * max_val)
            ub = 1.1 * max_val
            feature_vars[0][i] = scip.addVar(vtype="C", lb=lb, ub=ub, name=f"feature_{i}")

    amount_sold_vars[0][0] = scip.addVar(
        vtype="C", lb=0, ub=(1.1 * np.max(amount_sales)), name="amount_sold"
    )
    price_vars[0][0] = scip.addVar(
        vtype="C",
        lb=(max(0, np.min(price) - 0.1 * np.max(price))),
        ub=(1.1 * np.max(price)),
        name="price",
    )
    resale_vars[0][0] = scip.addVar(
        vtype="C",
        lb=(max(0, np.min(resale_price) - 0.1 * np.max(resale_price))),
        ub=(1.1 * np.max(resale_price)),
        name="resale_value",
    )

    # Create variables used for constraints that ensure the new car is sufficiently different from others
    pos_slack = np.zeros((n_features, n_car_comparisons), dtype=object)
    neg_slack = np.zeros((n_features, n_car_comparisons), dtype=object)
    binary_slack = np.zeros((n_features, n_car_comparisons), dtype=object)
    for i in range(1, n_features):
        for j in range(n_car_comparisons):
            big_m = feature_vars[0][i].getUbOriginal() - feature_vars[0][i].getLbOriginal()
            pos_slack[i][j] = scip.addVar(vtype="C", lb=0, ub=big_m, name=f"pos_slack_{i}_{j}")
            neg_slack[i][j] = scip.addVar(vtype="C", lb=0, ub=big_m, name=f"neg_slack_{i}_{j}")
            binary_slack[i][j] = scip.addVar(vtype="B", name=f"bin_slack_{i}_{j}")

    # Now add constraints to the model. Our constraints are that the new car is sufficiently different from those
    # already on the market, and that the resale value remains relatively high.
    top_selling_ids = np.argsort(price.reshape(-1))[-n_car_comparisons:]
    min_diff = data_random_state.uniform(low=2.8, high=3.3)
    for j, car_idx in enumerate(top_selling_ids):
        sum_slack = 0
        for i in range(1, n_features):
            big_m = feature_vars[0][i].getUbOriginal() - feature_vars[0][i].getLbOriginal()
            scip.addCons(
                pos_slack[i][j] <= big_m * binary_slack[i][j], name=f"bound_pos_slack_{i}_{j}"
            )
            scip.addCons(
                neg_slack[i][j] <= big_m * (1 - binary_slack[i][j]),
                name=f"bound_neg_slack_{i}_{j}",
            )
            feature_val = X[car_idx][i]
            scip.addCons(
                feature_vars[0][i] - feature_val == pos_slack[i][j] - neg_slack[i][j],
                name=f"l1_diff_{i}_{j}",
            )
            sum_slack += (pos_slack[i][j] + neg_slack[i][j]) / big_m
        scip.addCons(sum_slack >= min_diff, name=f"min_diff_{j}")

    # Now add the ML predictor constraints
    pred_cons = [
        add_predictor_constr(
            scip,
            reg_sales,
            np.concatenate((feature_vars, price_vars), axis=1),
            amount_sold_vars,
            epsilon=0.0001,
            formulation=formulation,
            unique_naming_prefix="num_sales_",
        ),
        add_predictor_constr(
            scip,
            reg_price,
            feature_vars,
            price_vars,
            epsilon=0.0001,
            formulation=formulation,
            unique_naming_prefix="price_",
        ),
        add_predictor_constr(
            scip,
            reg_resale,
            np.concatenate((feature_vars, price_vars), axis=1),
            resale_vars,
            epsilon=0.0001,
            formulation=formulation,
            unique_naming_prefix="resale_",
        ),
    ]

    # Add constraints that the resale value of the car must be sufficiently large and that the car is fuel efficient
    min_resale_ratio = data_random_state.uniform(low=0.70, high=0.75)
    min_fuel_efficiency = data_random_state.uniform(low=38, high=43)
    scip.addCons(resale_vars[0][0] >= min_resale_ratio * price_vars[0][0], name="min_resale")
    scip.addCons(feature_vars[0][8] >= min_fuel_efficiency, name="fuel_efficient")

    scip.setObjective(-amount_sold_vars[0][0] + 10000)

    if not build_only:
        # Optimise the SCIP model
        scip.optimize()
        # We can check the "error" of the MIP embedding by determining the difference SKLearn and SCIP output
        for predictor_constraint in pred_cons:
            if np.max(predictor_constraint.get_error()) > 10**-4:
                error = np.max(predictor_constraint.get_error())
                raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -4}")

    return scip


def test_gbdt_auto_manufacturer():
    scip = build_and_optimise_auto_manufacturer(
        training_seed=42,
        data_seed=42,
        gbdt_rf_or_mlp="gbdt",
        max_depth_or_layer_size=4,
        n_estimators_or_layers=4,
        build_only=False,
    )


def test_rf_auto_manufacturer():
    scip = build_and_optimise_auto_manufacturer(
        training_seed=42,
        data_seed=42,
        gbdt_rf_or_mlp="rf",
        max_depth_or_layer_size=4,
        n_estimators_or_layers=4,
        build_only=False,
    )


def test_mlp_auto_manufacturer():
    scip = build_and_optimise_auto_manufacturer(
        training_seed=42,
        data_seed=42,
        gbdt_rf_or_mlp="mlp",
        max_depth_or_layer_size=4,
        n_estimators_or_layers=2,
        build_only=False,
    )


def test_mlp_auto_manufacturer_bigm():
    scip = build_and_optimise_auto_manufacturer(
        training_seed=42,
        data_seed=42,
        gbdt_rf_or_mlp="mlp",
        formulation="bigm",
        max_depth_or_layer_size=4,
        n_estimators_or_layers=2,
        build_only=False,
    )



"""
PySCIPOpt Machine learning
==================================

A Python package to help automatically formulate and embed trained machine learning models in
mathematical optimization models. The package supports a variety of regression / classification machine learning models
(linear, logistic, neural networks, decision trees, random forests,...) trained by
different machine learning frameworks (scikit-learn, PyTorch, LightGBM, XGBoost).

See #TODO: Insert readthedocs link here
for documentation.
"""

# read version from installed package

from ._version import __version__
from .add_predictor import add_predictor_constr
from .registered_predictors import sklearn_convertors



"""Module for inserting ordinary Scikit-Learn regression models into a PySCIPOpt `Model`.

The following linear models are tested and should work:
   - :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
   - :external+sklearn:py:class:`sklearn.linear_model.Ridge`
   - :external+sklearn:py:class:`sklearn.linear_model.Lasso`
   - :external+sklearn:py:class:`sklearn.linear_model.ElasticNet`

"""

from .base_regression import BaseSKlearnRegressionConstr


def add_linear_regression_constr(
    scip_model, linear_regression, input_vars, output_vars=None, unique_naming_prefix="", **kwargs
):
    """Formulate linear_regression as a constraint in a PySCIPOpt Model.

    The formulation predicts the values of output_vars using input_vars according to
    linear_regression.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The scip model where the predictor should be inserted.
    linear_regression : :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
    The linear regression to insert. It can be any of the following types:
         * :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
         * :external+sklearn:py:class:`sklearn.linear_model.Ridge`
         * :external+sklearn:py:class:`sklearn.linear_model.Lasso`
         * :external+sklearn:py:class:`sklearn.linear_model.ElasticNet`
    input_vars: list or dict
        Decision variables used as input for the regression model
    output_vars: list or dict
        Decision variables used as output for the regression model
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    LinearRegressionConstr
        Object containing information about what was added to scip to formulate linear_regression.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return LinearRegressionConstr(
        scip_model, linear_regression, input_vars, output_vars, unique_naming_prefix, **kwargs
    )


class LinearRegressionConstr(BaseSKlearnRegressionConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.linear_model.LinearRegression` with SCIP

    |ClassShort|.
    """

    def __init__(
        self,
        scip_model,
        predictor,
        input_vars,
        output_vars=None,
        unique_naming_prefix="",
        **kwargs,
    ):
        if predictor.coef_.ndim > 1:
            self.output_size = predictor.coef_.shape[0]
        else:
            self.output_size = 1
        BaseSKlearnRegressionConstr.__init__(
            self,
            scip_model,
            predictor,
            input_vars,
            output_vars,
            unique_naming_prefix,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to SCIP."""
        self.add_regression_constr()



from warnings import warn

from .exceptions import NotRegistered
from .modelling.get_convertor import get_convertor
from .registered_predictors import registered_predictors


def add_predictor_constr(
    scip_model, predictor, input_vars, output_vars=None, unique_naming_prefix="p_", **kwargs
):
    """Formulate predictor in PySCIPOpt model.

    The formulation predicts the values of output_vars using input_vars according to
    predictor.

    Parameters
    ----------
    scip_model : PySCIPOpt Model
        The pyscipopt model where the predictor should be inserted.
    predictor:
        The predictor to insert.
    input_vars : list or np.ndarray
        Decision variables used as input for predictor in scip_model.
    output_vars : list or np.ndarray, optional
        Decision variables used as output for predictor in scip_model.
    unique_naming_prefix : str, optional
        A unique naming prefix that is used before all variable and constraint names. This parameter is important if
        the SCIP model is later printed to file and many predictors are added to the same SCIP model.

    Returns
    -------
    AbstractPredictorConstr
        Object containing information about what was added to scip_model to insert the
        predictor in it

    Note
    ----
    The parameters `input_vars` and `output_vars` can be either

     * Lists of variables (List of lists etc. for higher dimensional input)
     * np.ndarray of variables

     For internal use in the package they are cast into a np.ndarray of variables

    They should have dimensions that conform with the input/output of the predictor.
    We denote by `n_samples` the number of samples (or objects) that we want to predict with our predictor.
    We denote by `n_features` the dimension of the input of the predictor.
    We denote by `n_output` the dimension of the output.

    The `input_vars` are therefore of shape `(n_samples, n_features)` and the `output_vars` of
    shape `(n_samples, n_outputs)`. In the case of `output_vars` not being passed, appropriate variables will
    be automatically created.
    In the case of `n_samples == 1` the first dimension can simply be removed from the input.
    """
    convertors = registered_predictors()
    convertor = get_convertor(predictor, convertors)
    if convertor is None:
        raise NotRegistered(type(predictor).__name__)
    if len(unique_naming_prefix) > 0 and unique_naming_prefix[0].isdigit():
        warn(
            f"Unique naming prefix {unique_naming_prefix} begins with a digit and is unsafe for printing LP files"
        )
    return convertor(
        scip_model, predictor, input_vars, output_vars, unique_naming_prefix, **kwargs
    )



import numpy as np
import onnx
import pytest
from lightgbm import LGBMClassifier, LGBMRegressor
from pyscipopt import Model
from sklearn.base import ClusterMixin
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.cross_decomposition import PLSCanonical, PLSRegression
from sklearn.datasets import (
    load_breast_cancer,
    load_diabetes,
    load_iris,
    make_regression,
)
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
)
from sklearn.multioutput import MultiOutputClassifier, MultiOutputRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    Binarizer,
    Normalizer,
    OneHotEncoder,
    PolynomialFeatures,
    StandardScaler,
)
from sklearn.svm import SVC, SVR, LinearSVC, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from tensorflow import keras
from xgboost import XGBRegressor, XGBRFClassifier

from src.pyscipopt_ml import add_predictor_constr

"""
This is a set of tests for checking the basic errors of each formulation.
Single-output regression: Diabetes Dataset
Multi-output regression: Randomly generated dataset
Single-output classification: Breast Cancer Dataset
Multi-output classification: Iris Dataset
"""


def train_embed_and_optimise(predictor, multi_dimension, classification, n_samples=3, seed=42):
    # Create the training data
    if multi_dimension and classification:
        X, y = load_iris(return_X_y=True)
        input_size = 4
        output_size = 3
    elif multi_dimension and not classification:
        X, y = make_regression(
            n_samples=500, n_features=5, n_informative=5, n_targets=4, random_state=0, noise=0.5
        )
        input_size = 5
        output_size = 4
    elif not multi_dimension and classification:
        X, y = load_breast_cancer(return_X_y=True)
        input_size = 30
        output_size = 1
    else:
        X, y = load_diabetes(return_X_y=True)
        input_size = 10
        output_size = 1

    # Train the ML predictor
    if isinstance(predictor, keras.Model):
        # Define input
        predictor = keras.Sequential()
        predictor.add(keras.Input(shape=(input_size,)))

        # Add dense layers
        predictor.add(keras.layers.Dense(units=10, activation="relu"))
        predictor.add(keras.layers.Dense(units=10, activation="sigmoid"))
        predictor.add(keras.layers.Dense(units=output_size, activation="linear"))

        # Define output
        if classification and multi_dimension:
            predictor.add(keras.layers.Activation(keras.activations.softmax))
        predictor.compile(optimizer="adam", loss="mse")
    if isinstance(predictor, onnx.ModelProto):
        pass
    elif not isinstance(predictor, ClusterMixin):
        if isinstance(predictor, MultiOutputClassifier) or (
            isinstance(predictor, keras.Model) and multi_dimension and classification
        ):
            y = OneHotEncoder(sparse_output=False).fit_transform(y.reshape(-1, 1))
        predictor.fit(X, y)
    else:
        predictor.fit(X)

    # Create the SCIP model
    scip = Model()

    # Create the input and output variables
    input_vars = np.zeros(shape=(n_samples, input_size), dtype=object)
    output_vars = np.zeros(shape=(n_samples, output_size), dtype=object)
    # Create the random variable fixings (from samples in training)
    rng = np.random.default_rng(seed=seed)
    random_integers = rng.integers(0, len(X), size=n_samples)
    for i in range(n_samples):
        for j in range(input_size):
            input_vars[i][j] = scip.addVar(name=f"x_{i}_{j}", vtype="C")
            scip.fixVar(input_vars[i][j], X[random_integers[i]][j] + 1e-04 * rng.random())
        for j in range(output_size):
            output_vars[i][j] = scip.addVar(name=f"y_{i}_{j}", vtype="C", lb=None)

    # Embed the trained model
    if (
        isinstance(predictor, keras.Model) or isinstance(predictor, onnx.ModelProto)
    ) and classification:
        pred_cons = add_predictor_constr(
            scip, predictor, input_vars, output_vars, output_type="classification"
        )
    else:
        pred_cons = add_predictor_constr(scip, predictor, input_vars, output_vars)

    # Optimise the model and check the error
    print(f"Optimising predictor {predictor}")
    # scip.hideOutput()
    scip.optimize()

    # Ensure the error is below some threshold
    if np.max(pred_cons.get_error()) > 10**-4:
        error = np.max(pred_cons.get_error())
        raise AssertionError(f"Max error {error} exceeds threshold of {10 ** -4}")


testdata = [
    (Pipeline([("scaler", StandardScaler()), ("linreg", LinearRegression())]), True, False),
    (LinearRegression(), True, False),
    (ElasticNet(), False, False),
    (Lasso(), True, False),
    (Ridge(), False, False),
    (PLSCanonical(n_components=4), True, False),
    (PLSRegression(n_components=1), False, False),
    (Pipeline([("s", StandardScaler()), ("l", LogisticRegression())]), False, True),
    (Pipeline([("s", StandardScaler()), ("l", LogisticRegression())]), True, True),
    (Pipeline([("b", Binarizer()), ("d", DecisionTreeRegressor(max_depth=4))]), False, False),
    (DecisionTreeRegressor(max_depth=4), True, False),
    (DecisionTreeRegressor(max_depth=4), False, False),
    (DecisionTreeClassifier(max_depth=4), True, True),
    (DecisionTreeClassifier(max_depth=4), False, True),
    (GradientBoostingRegressor(max_depth=4, n_estimators=4), False, False),
    (MultiOutputRegressor(GradientBoostingRegressor(max_depth=4, n_estimators=4)), True, False),
    (GradientBoostingClassifier(max_depth=4, n_estimators=4), True, True),
    (MultiOutputClassifier(GradientBoostingClassifier(max_depth=4, n_estimators=4)), True, True),
    (RandomForestRegressor(max_depth=4, n_estimators=4), True, False),
    (RandomForestClassifier(max_depth=4, n_estimators=4), True, True),
    (RandomForestClassifier(max_depth=4, n_estimators=4), False, True),
    (onnx.load("./tests/data/sk_3_42_0_0_relu.onnx"), False, False),
    (MLPRegressor((10, 10, 10), "relu"), False, False),
    (onnx.load("./tests/data/sk_3_42_1_0_logistic.onnx"), True, False),
    (MLPRegressor((10, 10, 10), "logistic"), True, False),
    (onnx.load("./tests/data/sk_3_42_1_0_tanh.onnx"), True, False),
    (MLPRegressor((10, 10, 10), "tanh"), True, False),
    (MLPClassifier((10, 10, 10), "relu"), False, True),
    (MLPClassifier((10, 10, 10), "relu"), True, True),
    (MLPClassifier((10, 10, 10), "logistic"), True, True),
    (LinearSVR(dual="auto"), False, False),
    (MultiOutputRegressor(LinearSVR(dual="auto")), True, False),
    (SVR(kernel="poly"), False, False),
    (SVR(kernel="linear"), False, False),
    (LinearSVC(dual="auto", max_iter=5000), False, True),
    (SVC(kernel="poly", degree=2), False, True),
    (MultiOutputClassifier(SVC(kernel="poly", degree=2)), True, True),
    (XGBRegressor(max_depth=4, n_estimators=4), True, False),
    (XGBRegressor(max_depth=4, n_estimators=4), False, False),
    (XGBRFClassifier(max_depth=4, n_estimators=4), False, True),
    (LGBMRegressor(max_depth=4, n_estimators=4), False, False),
    (LGBMClassifier(max_depth=4, n_estimators=4), True, True),
    (LGBMClassifier(max_depth=4, n_estimators=4), False, True),
    (keras.Model(), True, False),
    (keras.Model(), True, True),
    (onnx.load("./tests/data/keras_3_42_1_0.onnx"), True, False),
    (onnx.load("./tests/data/keras_3_42_1_1.onnx"), True, True),
    (KMeans(n_clusters=3, n_init="auto"), True, True),
    (KMeans(n_clusters=2, n_init="auto"), False, True),
    (MiniBatchKMeans(n_clusters=3, n_init="auto"), True, True),
    (Pipeline([("s", PolynomialFeatures(degree=3)), ("l", LinearRegression())]), True, False),
    (Pipeline([("n", Normalizer(norm="l1")), ("l", LinearRegression())]), True, False),
    (Pipeline([("n", Normalizer(norm="l2")), ("l", LinearRegression())]), True, False),
    (Pipeline([("n", Normalizer(norm="max")), ("l", LinearRegression())]), True, False),
    (Pipeline([("b", Binarizer()), ("l", LinearRegression())]), True, False),
]


@pytest.mark.parametrize("predictor,multi_dimension,classification", testdata)
def test_formulation_error(predictor, multi_dimension, classification):
    train_embed_and_optimise(predictor, multi_dimension, classification)
